import time
from collections import defaultdict
from typing import Optional

from vllm.utils import logger, sha256
from vllm.v1.core.block_pool import BlockPool
from vllm.v1.core.kv_cache_utils import (BlockHash, KVCacheBlock,
                                         PrefixCachingMetrics)
from vllm.v1.core.single_type_kv_cache_manager import \
    get_manager_for_kv_cache_spec
from vllm.v1.kv_cache_interface import KVCacheSpec
from vllm.v1.metrics.stats import PrefixCacheStats
from vllm.v1.request import Request


class CPUCacheStats:

    def __init__(self, enable_prefix_caching: bool, log_stats: bool = False):
        self.enable_prefix_caching = enable_prefix_caching
        self.log_stats = log_stats
        self.prefix_cache_stats = PrefixCacheStats() if log_stats else None
        self.cpu_prefix_cache_metrics = PrefixCachingMetrics()
        self.time_sec = int(time.time())

    def log(self):
        current_time_sec = int(time.time())
        # Log the prefix cache hit rate every 10 seconds.
        if current_time_sec - self.time_sec >= 10:
            self.time_sec = current_time_sec
            logger.info("CPU Prefix cache hit rate: %.1f%%",
                        self.cpu_prefix_cache_metrics.hit_rate * 100)

    def make_prefix_cache_stats(self) -> Optional[PrefixCacheStats]:
        """Get (and reset) the prefix cache stats.
        Returns:
            The current prefix caching stats, or None if logging is disabled.
        """
        if not self.log_stats:
            return None
        stats = self.prefix_cache_stats
        self.prefix_cache_stats = PrefixCacheStats()
        return stats

    def update(self, num_tokens, num_computed_tokens):
        # Note the function is called by scheduler
        if self.log_stats and self.enable_prefix_caching:
            assert self.prefix_cache_stats is not None
            self.prefix_cache_stats.requests += 1
            self.prefix_cache_stats.queries += num_tokens
            self.prefix_cache_stats.hits += num_computed_tokens

    def set_cache_stats(self, num_tokens, num_computed_tokens):
        assert self.prefix_cache_stats is not None
        self.prefix_cache_stats.hits = num_computed_tokens
        self.prefix_cache_stats.queries = num_tokens
        self.prefix_cache_stats.requests = 1


class CPUKVCacheManager:

    def __init__(
        self,
        kv_cache_spec: KVCacheSpec,
        num_cpu_blocks: int,
        caching_hash_algo: str = "builtin",
        use_eagle: bool = False,
        enable_kv_cache_events: bool = False,
    ) -> None:
        self.block_size = kv_cache_spec.block_size
        self.num_cpu_blocks = num_cpu_blocks
        self.caching_hash_fn = sha256 if caching_hash_algo == "sha256" else hash
        self.use_eagle = use_eagle
        self.block_pool = BlockPool(self.num_cpu_blocks, True,
                                    enable_kv_cache_events)
        self.single_type_manager = get_manager_for_kv_cache_spec(
            kv_cache_spec=kv_cache_spec,
            block_pool=self.block_pool,
            kv_cache_group_id=0,
        )
        # Record kv block hashes, avoid redundant computation.
        self.req_to_block_hashes: defaultdict[
            str, list[BlockHash]] = defaultdict(list)
        # Record blocks touched in get_matched_num_and_touch().
        self.req_to_computed_blocks: defaultdict[
            str, list[KVCacheBlock]] = defaultdict(list)
        # Record the request that failed to allocate.
        self.req_failed_to_allocate: defaultdict[str, bool] = defaultdict(bool)
        self.req_to_num_tokens: defaultdict[str, int] = defaultdict(int)
        self.cpu_cache_stats = CPUCacheStats(enable_prefix_caching=True,
                                             log_stats=True)
        # Record request that will be free after finish sending
        self.req_to_free: defaultdict[str, Request] = defaultdict(Request)

    def get_matched_num_and_touch(self, request: Request) -> tuple[int, bool]:
        # When the request requires prompt logprobs, we skip prefix caching.
        if (request.sampling_params.prompt_logprobs is not None):
            return 0, False
        request_id = request.request_id
        # The block hashes for the request may already be computed
        # if the scheduler has tried to schedule the request before.
        block_hashes = self.req_to_block_hashes[request_id]
        if not block_hashes:
            block_hashes = request.block_hashes
            self.req_to_block_hashes[request_id] = block_hashes
        max_cache_hit_length = request.num_tokens - 1
        computed_blocks = self.single_type_manager.find_longest_cache_hit(
            block_hashes=block_hashes,
            max_length=max_cache_hit_length,
            kv_cache_group_ids=[0],
            block_pool=self.block_pool,
            kv_cache_spec=self.single_type_manager.kv_cache_spec,
            use_eagle=self.use_eagle,
        )
        num_computed_tokens = len(computed_blocks[0]) * self.block_size
        self.req_to_computed_blocks[request_id] = computed_blocks[0]
        # We should touch these blocks in the concurrent scenarios.
        self.block_pool.touch(computed_blocks)

        # cup prefix cache status set and log
        assert self.cpu_cache_stats is not None and self.cpu_cache_stats.prefix_cache_stats is not None
        self.cpu_cache_stats.set_cache_stats(request.num_tokens,
                                             num_computed_tokens)
        self.cpu_cache_stats.cpu_prefix_cache_metrics.observe(
            self.cpu_cache_stats.prefix_cache_stats)
        self.cpu_cache_stats.log()

        return num_computed_tokens, False

    def _release_ahead_touch(self, request_id: str):
        computed_blocks = self.req_to_computed_blocks[request_id]
        if computed_blocks:
            self.single_type_manager.block_pool.free_blocks(
                reversed(computed_blocks))
            self.req_to_computed_blocks.pop(request_id, None)

    def allocate_slots(self, req_to_num_tokens: dict[str, int],
                       unallocated_req_ids: set[str]) -> dict[str, list[int]]:
        for request_id in unallocated_req_ids:
            self._free_slots(request_id)
        req_to_new_blocks = {}
        for request_id, num_tokens in req_to_num_tokens.items():
            if self.req_failed_to_allocate[request_id]:
                continue
            new_computed_blocks = self.req_to_computed_blocks[request_id]
            num_blocks_to_allocate = (
                self.single_type_manager.get_num_blocks_to_allocate(
                    request_id=request_id,
                    num_tokens=num_tokens,
                    new_computed_blocks=new_computed_blocks,
                ))
            if num_blocks_to_allocate > self.block_pool.get_num_free_blocks():
                self._release_ahead_touch(request_id)
                self.req_failed_to_allocate[request_id] = True
                continue
            # Append the new computed blocks to the request blocks until now to
            # avoid the case where the new blocks cannot be allocated.
            self.single_type_manager.save_new_computed_blocks(
                request_id, new_computed_blocks)
            # Allocate new blocks but do not cache now.
            new_blocks = self.single_type_manager.allocate_new_blocks(
                request_id, num_tokens)
            self.req_to_num_tokens[request_id] = num_tokens
            # No need to release ref_cnt because we use officially.
            self.req_to_computed_blocks.pop(request_id, None)
            req_to_new_blocks[request_id] = [
                block.block_id for block in new_computed_blocks + new_blocks
            ]
        return req_to_new_blocks

    def record_request_cache_and_free_slots(self, request: Request):
        logger.debug(
            f"record_request_cache_and_free_slots for request {request.request_id} in cpu_kv_cache_manager"
        )
        self.req_to_free[request.request_id] = request

    def cache_and_free_slots(self, request_id: str):
        logger.debug(
            f"Cache and free slots for request {request_id} in cpu_kv_cache_manager"
        )
        if request_id not in self.req_to_free:
            logger.Error(
                f"request {request_id} not in req_to_free, maybe bug!")
            return
        request = self.req_to_free[request_id]
        if not self.req_failed_to_allocate[request_id]:
            self.single_type_manager.cache_blocks(
                request,
                self.req_to_num_tokens[request_id],
            )
        self._free_slots(request_id)
        logger.debug(
            f"delete request {request_id} in cpu_kv_cache_manager req_to_free")
        del self.req_to_free[request_id]

    def _free_slots(self, request_id: str):
        # This function is designed to be reentrant.
        self._release_ahead_touch(request_id)
        self.single_type_manager.free(request_id)
        self.req_to_block_hashes.pop(request_id, None)
        self.req_to_computed_blocks.pop(request_id, None)
        self.req_failed_to_allocate.pop(request_id, None)
        self.req_to_num_tokens.pop(request_id, None)



from typing import Optional, Union

import numpy as np
import torch
from vllm.distributed import get_dcp_group
from vllm.utils import cdiv


class BlockTable:

    def __init__(self,
                 block_size: int,
                 max_num_reqs: int,
                 max_num_blocks_per_req: int,
                 max_num_batched_tokens: int,
                 pin_memory: bool,
                 device: torch.device,
                 kernel_sizes: Union[list[int], None] = None):
        self.max_num_reqs = max_num_reqs
        self.max_num_blocks_per_req = max_num_blocks_per_req
        self.max_num_batched_tokens = max_num_batched_tokens
        self.pin_memory = pin_memory
        self.device = device
        self.physical_block_size = block_size
        # If kernel_sizes is None or [0], use physical block size (no splitting)
        if kernel_sizes is None or kernel_sizes == [0]:
            self.block_size = block_size
            self.logical_block_size = block_size
            self.blocks_per_phys_block = 1
            self.use_hybrid_blocks = False
        else:
            # Find the first kernel size that divides physical_block_size evenly
            selected_kernel_size = None
            for kernel_size in kernel_sizes:
                if kernel_size > 0 \
                    and self.physical_block_size % kernel_size == 0:
                    selected_kernel_size = kernel_size
                    break

            if selected_kernel_size is None:
                raise ValueError(
                    f"None of the kernel sizes {kernel_sizes} can divide "
                    f"physical block size {self.physical_block_size} evenly")

            self.block_size = selected_kernel_size
            self.logical_block_size = selected_kernel_size
            self.blocks_per_phys_block = (self.physical_block_size //
                                          self.logical_block_size)
            if self.blocks_per_phys_block > 1:
                self.use_hybrid_blocks = True
            else:
                self.use_hybrid_blocks = False

        if self.use_hybrid_blocks:
            logical_table_size = (max_num_blocks_per_req *
                                  self.blocks_per_phys_block)
        else:
            logical_table_size = max_num_blocks_per_req

        self.block_table = torch.zeros(
            (max_num_reqs, logical_table_size),
            device=self.device,
            dtype=torch.int32,
        )
        self.block_table_cpu = torch.zeros(
            (max_num_reqs, logical_table_size),
            device="cpu",
            dtype=torch.int32,
            pin_memory=pin_memory,
        )
        self.block_table_np = self.block_table_cpu.numpy()
        self.num_blocks_per_row = np.zeros(max_num_reqs, dtype=np.int32)

        self.slot_mapping_cpu = torch.zeros(self.max_num_batched_tokens,
                                            dtype=torch.int64,
                                            device="cpu",
                                            pin_memory=self.pin_memory)
        self.slot_mapping_np = self.slot_mapping_cpu.numpy()
        self.slot_mapping = torch.zeros(self.max_num_batched_tokens,
                                        dtype=torch.int64,
                                        device=self.device)
        try:
            self.dcp_world_size = get_dcp_group().world_size
            self.dcp_rank = get_dcp_group().rank_in_group
        except AssertionError:
            # DCP might not be initialized in testing
            self.dcp_world_size = 1
            self.dcp_rank = 0
        self.kernel_sizes = kernel_sizes

    def append_row(
        self,
        block_ids,
        row_idx: int,
    ) -> None:
        if not block_ids:
            return
        block_ids = np.array(block_ids)
        if self.use_hybrid_blocks:
            block_ids = self._convert_physical_to_logical_blocks(block_ids)

        num_blocks = len(block_ids)
        start = self.num_blocks_per_row[row_idx]

        self.block_table_np[row_idx, start:start + num_blocks] = block_ids
        self.num_blocks_per_row[row_idx] += num_blocks

    def add_row(self, block_ids: list[int], row_idx: int) -> None:
        self.num_blocks_per_row[row_idx] = 0
        self.append_row(block_ids, row_idx)

    def move_row(self, src: int, tgt: int) -> None:
        num_blocks = self.num_blocks_per_row[src]
        self.block_table_np[tgt, :num_blocks] = self.block_table_np[
            src, :num_blocks]
        self.num_blocks_per_row[tgt] = num_blocks

    def swap_row(self, src: int, tgt: int) -> None:
        num_blocks_src = self.num_blocks_per_row[src]
        num_blocks_tgt = self.num_blocks_per_row[tgt]
        self.num_blocks_per_row[src] = num_blocks_tgt
        self.num_blocks_per_row[tgt] = num_blocks_src

        self.block_table_np[[src, tgt]] = self.block_table_np[[tgt, src]]

    def compute_slot_mapping(self, req_indices: np.ndarray,
                             positions: np.ndarray) -> None:
        # E.g., [0, 1, 0, 1, 2, 3, 4, 0, 1, 2]
        # -> [0, 0, K, K, K + 1, K + 1, K + 2, 2 * K, 2 * K, 2 * K + 1]
        # where K is the max_num_blocks_per_req and the block size is 2.
        # NOTE(woosuk): We can't simply use `token_indices // block_size`
        # here because M (max_model_len) is not necessarily divisible by
        # block_size.

        if self.dcp_world_size > 1:
            # Note(hc): The DCP implement store kvcache with an interleave
            # style, the kvcache for the token whose token_idx is i is
            # always stored on the GPU whose dcp_rank equals i % cp_world_size:

            # Use a "virtual block" which equals to world_size * block_size
            # for block_table_indices calculation.
            virtual_block_size = self.block_size * self.dcp_world_size

            # IMPORTANT: In hybrid mode, positions are in logical block space,
            # but we need to map them to the correct logical block table indices
            logical_block_idx = positions // virtual_block_size

            # Account for the expanded logical table
            # (always needed with unified tensor)
            # Each physical block is split into multiple logical blocks
            # The logical table has been expanded to accommodate this
            block_table_indices = (req_indices * self.max_num_blocks_per_req *
                                   self.blocks_per_phys_block +
                                   logical_block_idx)

            block_numbers = self.block_table_np.ravel()[block_table_indices]
            # Use virtual_block_size for mask calculation, which marks local
            # tokens.
            virtual_block_offsets = positions % virtual_block_size
            mask = virtual_block_offsets % self.dcp_world_size == self.dcp_rank
            # Calculate local block_offsets
            block_offsets = virtual_block_offsets // self.dcp_world_size
            # Calculate slot_mapping
            slot_mapping = block_numbers * self.block_size + block_offsets
            # Write final slots, use -1 for not-local
            self.slot_mapping_np[:req_indices.shape[0]] = np.where(
                mask, slot_mapping, -1)
        else:
            assert self.kernel_sizes is not None
            if self.block_size == self.kernel_sizes[0]:
                # IMPORTANT: In hybrid mode, positions are in logical block space,
                # but we need to map them to the correct logical block table indices
                logical_block_idx = positions // self.block_size

                # Account for the expanded logical table
                # (always needed with unified tensor)
                # Each physical block is split into multiple logical blocks
                # The logical table has been expanded to accommodate this
                block_table_indices = (
                    req_indices * self.max_num_blocks_per_req *
                    self.blocks_per_phys_block + logical_block_idx)

                block_numbers = self.block_table_np.ravel(
                )[block_table_indices]
                block_offsets = positions % self.block_size
                np.add(block_numbers * self.block_size,
                       block_offsets,
                       out=self.slot_mapping_np[:req_indices.shape[0]])

    def commit_block_table(self, num_reqs: int) -> None:
        self.block_table[:num_reqs].copy_(self.block_table_cpu[:num_reqs],
                                          non_blocking=True)

    def commit_slot_mapping(self, num_tokens: int) -> None:
        self.slot_mapping[:num_tokens].copy_(
            self.slot_mapping_cpu[:num_tokens], non_blocking=True)

    def clear(self) -> None:
        self.block_table.fill_(0)
        self.block_table_cpu.fill_(0)

    def _convert_physical_to_logical_blocks(
            self, physical_blocks: np.ndarray) -> np.ndarray:
        """Convert physical block IDs to logical block IDs."""
        if not self.use_hybrid_blocks:
            return physical_blocks

        # Create logical block IDs by splitting each physical block
        logical_blocks: list[int] = []
        for phys_block in physical_blocks:
            # Convert physical block to multiple logical blocks
            # Physical block 1 becomes logical blocks
            # [1*split_ratio, 1*split_ratio+1, ...]
            # But we need to account for the fact that block 0 is special
            base_logical = phys_block * self.blocks_per_phys_block
            logical_blocks.extend(
                range(base_logical, base_logical + self.blocks_per_phys_block))

        return np.array(logical_blocks, dtype=np.int32)

    def get_device_tensor(self) -> torch.Tensor:
        """Returns the device tensor of the block table."""
        return self.block_table

    def get_cpu_tensor(self) -> torch.Tensor:
        """Returns the CPU tensor of the block table."""
        return self.block_table_cpu

    def get_numpy_array(self) -> np.ndarray:
        """Returns the numpy array of the block table."""
        return self.block_table_np


class MultiGroupBlockTable:
    """The BlockTables for each KV cache group."""

    def __init__(self,
                 max_num_reqs: int,
                 max_model_len: int,
                 max_num_batched_tokens: int,
                 pin_memory: bool,
                 device: torch.device,
                 block_sizes: list[int],
                 num_speculative_tokens: int = 0,
                 kernel_sizes: Optional[list[list[int]]] = None) -> None:
        # Note(hc): each dcp rank only store
        # (max_model_len//dcp_world_size) tokens in kvcache,
        # so the block_size which used for calc max_num_blocks_per_req
        # must be multiplied by dcp_world_size.
        try:
            dcp_world_size = get_dcp_group().world_size
        except AssertionError:
            # DCP might not be initialized in testing
            dcp_world_size = 1

        if kernel_sizes is None:
            kernel_sizes = [[0]] * len(block_sizes)
        # Ensure kernel_sizes matches block_sizes length
        elif len(kernel_sizes) == 1 and len(block_sizes) > 1:
            kernel_sizes = kernel_sizes * len(block_sizes)
        elif len(kernel_sizes) != len(block_sizes):
            raise ValueError(
                f"kernel_sizes length ({len(kernel_sizes)}) must match "
                f"block_sizes length ({len(block_sizes)})")

        # Use zip to pair block_sizes with kernel_sizes one-to-one
        self.block_tables = [
            BlockTable(
                block_size, max_num_reqs,
                max(cdiv(max_model_len, block_size * dcp_world_size),
                    1 + num_speculative_tokens), max_num_batched_tokens,
                pin_memory, device, kernel_size_list)
            for block_size, kernel_size_list in zip(block_sizes, kernel_sizes)
        ]

    def append_row(self, block_ids: tuple[list[int], ...],
                   row_idx: int) -> None:
        for i, block_table in enumerate(self.block_tables):
            block_table.append_row(block_ids[i], row_idx)

    def add_row(self, block_ids: tuple[list[int], ...], row_idx: int) -> None:
        for i, block_table in enumerate(self.block_tables):
            block_table.add_row(block_ids[i], row_idx)

    def move_row(self, src: int, tgt: int) -> None:
        for block_table in self.block_tables:
            block_table.move_row(src, tgt)

    def swap_row(self, src: int, tgt: int) -> None:
        for block_table in self.block_tables:
            block_table.swap_row(src, tgt)

    def compute_slot_mapping(self, req_indices: np.ndarray,
                             positions: np.ndarray) -> None:
        for block_table in self.block_tables:
            block_table.compute_slot_mapping(req_indices, positions)

    def commit_block_table(self, num_reqs: int) -> None:
        for block_table in self.block_tables:
            block_table.commit_block_table(num_reqs)

    def commit_slot_mapping(self, num_tokens: int) -> None:
        for block_table in self.block_tables:
            block_table.commit_slot_mapping(num_tokens)

    def clear(self) -> None:
        for block_table in self.block_tables:
            block_table.clear()

    def __getitem__(self, idx: int) -> "BlockTable":
        """Returns the BlockTable for the i-th KV cache group."""
        return self.block_tables[idx]



import math
import unittest
from unittest.mock import MagicMock, PropertyMock, patch

import torch
from transformers.configuration_utils import PretrainedConfig
from vllm.config import ModelConfig, VllmConfig
from vllm.model_executor.layers.rotary_embedding import (
    DeepseekScalingRotaryEmbedding, MRotaryEmbedding, RotaryEmbedding)
from vllm.platforms import CpuArchEnum

from tests.ut.base import TestBase
from vllm_ascend.ascend_forward_context import set_ascend_forward_context
from vllm_ascend.ops.rotary_embedding import _custom_rotary_embedding_enabled

MODEL = "Qwen3-0.6B"
MODEL_VL = "Qwen/Qwen2.5-VL-3B-Instruct"
MAX_NUM_BATCHED_TOKEND = 10000


class TestCustomRotaryEmbeddingEnabled(unittest.TestCase):

    def setUp(self):
        # Common setup for tests
        self.positions = torch.tensor([1, 2, 3])
        self.query = torch.randn(3, 4, dtype=torch.float16)
        self.key = torch.randn(3, 4, dtype=torch.float16)
        self.head_size = 32
        self.cos_sin_cache = torch.randn(3, 4)

        # Mock self object for rope_forward_oot
        self.mock_self = MagicMock()
        self.mock_self.head_size = self.head_size
        self.mock_self.cos_sin_cache = self.cos_sin_cache
        self.mock_self.is_neox_style = True
        self.mock_self.forward_native.return_value = (self.query, self.key)

    def test_custom_rotary_embedding_enabled(self):
        # Test when all conditions are True
        with patch('vllm_ascend.ops.rotary_embedding.enable_custom_op',
                   return_value=True):
            result = _custom_rotary_embedding_enabled(self.query, True,
                                                      self.head_size)
            self.assertTrue(result)

        # Test when dtype is not float16
        with patch('vllm_ascend.ops.rotary_embedding.enable_custom_op',
                   return_value=True):
            query = self.query.to(torch.float32)
            result = _custom_rotary_embedding_enabled(query, True,
                                                      self.head_size)
            self.assertFalse(result)

        # Test when neox_style is False
        with patch('vllm_ascend.ops.rotary_embedding.enable_custom_op',
                   return_value=True):
            result = _custom_rotary_embedding_enabled(self.query, False,
                                                      self.head_size)
            self.assertFalse(result)

        # Test when head_size is not divisible by 32
        with patch('vllm_ascend.ops.rotary_embedding.enable_custom_op',
                   return_value=True):
            result = _custom_rotary_embedding_enabled(self.query, True,
                                                      self.head_size + 1)
            self.assertFalse(result)

        # Test when custom op is disabled
        with patch('vllm_ascend.ops.rotary_embedding.enable_custom_op',
                   return_value=False):
            result = _custom_rotary_embedding_enabled(self.query, True,
                                                      self.head_size)
            self.assertFalse(result)


class TestAscendRotaryEmbedding(unittest.TestCase):

    def setUp(self):
        # Common setup for tests
        self.positions = torch.tensor([1, 2, 3])
        self.query = torch.randn(3, 1, 32, dtype=torch.float16)
        self.key = torch.randn(3, 1, 32, dtype=torch.float16)
        self.head_size = 32
        self.rotary_dim = self.head_size
        self.max_position = 16
        self.rope_theta = 10000
        self.is_neox_style = True
        self.cos_sin_cache = torch.randn(3, 1, 32)
        self.layer = RotaryEmbedding(self.head_size, self.rotary_dim,
                                     self.max_position, self.rope_theta,
                                     self.is_neox_style, torch.float16)

        # Mock self object for rope_forward_oot
        self.mock_self = MagicMock()
        self.mock_self.head_size = self.head_size
        self.mock_self.cos_sin_cache = self.cos_sin_cache
        self.mock_self.is_neox_style = self.is_neox_style

    @patch('torch.ops._C_ascend')
    @patch('vllm_ascend.ops.rotary_embedding.is_310p', return_value=False)
    @patch('vllm_ascend.ops.rotary_embedding._custom_rotary_embedding_enabled',
           return_value=True)
    @patch('torch.ops._npu_rotary_embedding')
    @patch('vllm.config.ModelConfig.__post_init__', MagicMock())
    @patch('vllm.config.VllmConfig.__post_init__', MagicMock())
    @patch('vllm.distributed.parallel_state._DP', MagicMock(world_size=1))
    @patch('vllm.distributed.parallel_state._TP', MagicMock(world_size=1))
    def test_rope_forward_oot_custom_kernel(self, mock_rotary_embedding,
                                            mock_custom_enabled, mock_is_310p,
                                            mock__c):
        mock_config = MagicMock()
        mock_config.torchair_graph_config.enabled = False

        # Setup mock for custom kernel path

        mock__c.rotary_embedding.return_value = self.query, self.key
        vllm_config = VllmConfig()
        model_config = ModelConfig(MODEL,
                                   tokenizer=MODEL,
                                   max_model_len=MAX_NUM_BATCHED_TOKEND)
        model_config.hf_config = PretrainedConfig()
        vllm_config.model_config = model_config
        with set_ascend_forward_context(None, vllm_config):
            result_q, result_k = self.layer.forward(self.positions, self.query,
                                                    self.key)

        mock__c.rotary_embedding.assert_called_once()
        self.assertEqual(result_q.shape, self.query.shape)
        self.assertEqual(result_k.shape, self.key.shape)

    @patch('vllm_ascend.ops.rotary_embedding._custom_rotary_embedding_enabled',
           return_value=False)
    @patch('torch_npu._npu_rotary_embedding')
    @patch('vllm.config.ModelConfig.__post_init__', MagicMock())
    @patch('vllm.config.VllmConfig.__post_init__', MagicMock())
    @patch('vllm.distributed.parallel_state._DP', MagicMock(world_size=1))
    @patch('vllm.distributed.parallel_state._TP', MagicMock(world_size=1))
    def test_rope_forward_oot_contiguous(self, mock_npu_rotary,
                                         mock_custom_enabled):
        mock_config = MagicMock()
        mock_config.torchair_graph_config.enabled = False

        # Test contiguous path when custom is disabled
        non_contig_query = self.query.transpose(0, 1)
        non_contig_key = self.key.transpose(0, 1)
        vllm_config = VllmConfig()
        model_config = ModelConfig(MODEL,
                                   tokenizer=MODEL,
                                   max_model_len=MAX_NUM_BATCHED_TOKEND)
        model_config.hf_config = PretrainedConfig()
        vllm_config.model_config = model_config
        with set_ascend_forward_context(None, vllm_config):
            result_q, result_k = self.layer.forward(self.positions,
                                                    non_contig_query,
                                                    non_contig_key)

        mock_npu_rotary.assert_called_once()
        self.assertEqual(result_q.shape, non_contig_query.shape)
        self.assertEqual(result_k.shape, non_contig_key.shape)

    @patch('vllm.config.ModelConfig.__post_init__', MagicMock())
    @patch('vllm.config.VllmConfig.__post_init__', MagicMock())
    @patch('vllm.distributed.parallel_state._DP', MagicMock(world_size=1))
    @patch('vllm.distributed.parallel_state._TP', MagicMock(world_size=1))
    def test_rope_forward_oot_with_offsets(self):
        mock_config = MagicMock()
        mock_config.torchair_graph_config.enabled = False

        # Test that NotImplementedError is raised when offsets is provided
        offsets = torch.tensor([1, 2, 3])
        with self.assertRaises(NotImplementedError):
            vllm_config = VllmConfig()
            model_config = ModelConfig(MODEL,
                                       tokenizer=MODEL,
                                       max_model_len=MAX_NUM_BATCHED_TOKEND)
            model_config.hf_config = PretrainedConfig()
            vllm_config.model_config = model_config
            with set_ascend_forward_context(None, vllm_config):
                self.layer.forward(self.positions, self.query, self.key,
                                   offsets)

    @patch('vllm_ascend.ops.rotary_embedding._custom_rotary_embedding_enabled',
           return_value=False)
    @patch('torch_npu._npu_rotary_embedding')
    @patch('vllm.config.ModelConfig.__post_init__', MagicMock())
    @patch('vllm.config.VllmConfig.__post_init__', MagicMock())
    @patch('vllm.distributed.parallel_state._DP', MagicMock(world_size=1))
    @patch('vllm.distributed.parallel_state._TP', MagicMock(world_size=1))
    def test_rope_forward_oot_neox_style_override(self, mock_npu_rotary,
                                                  mock_custom_enabled):
        mock_config = MagicMock()
        mock_config.torchair_graph_config.enabled = False

        # Test neox_style override
        vllm_config = VllmConfig()
        model_config = ModelConfig(MODEL,
                                   tokenizer=MODEL,
                                   max_model_len=MAX_NUM_BATCHED_TOKEND)
        model_config.hf_config = PretrainedConfig()
        vllm_config.model_config = model_config
        with set_ascend_forward_context(None, vllm_config):
            result_q, result_k = self.layer.forward(
                self.positions,
                self.query,
                self.key,
                is_neox_style_override=False)
        # Check that neox_style=False was passed to the NPU function
        args, kwargs = mock_npu_rotary.call_args
        self.assertFalse(args[-1])

    @patch('vllm_ascend.ops.rotary_embedding._custom_rotary_embedding_enabled',
           return_value=False)
    @patch('torch_npu._npu_rotary_embedding')
    @patch('vllm.config.ModelConfig.__post_init__', MagicMock())
    @patch('vllm.config.VllmConfig.__post_init__', MagicMock())
    @patch('vllm.distributed.parallel_state._DP', MagicMock(world_size=1))
    @patch('vllm.distributed.parallel_state._TP', MagicMock(world_size=1))
    def test_rope_forward_oot_rotary_dim_less_than_head_size(
            self, mock_npu_rotary, mock_custom_enabled):
        mock_config = MagicMock()
        mock_config.torchair_graph_config.enabled = False

        # test case when rotary_dim < head_size
        org_rotary_dim = self.layer.rotary_dim
        self.layer.rotary_dim = self.layer.head_size // 2

        vllm_config = VllmConfig()
        model_config = ModelConfig(MODEL,
                                   tokenizer=MODEL,
                                   max_model_len=MAX_NUM_BATCHED_TOKEND)
        model_config.hf_config = PretrainedConfig()
        vllm_config.model_config = model_config
        with set_ascend_forward_context(None, vllm_config):
            result_q, result_k = self.layer.forward(self.positions, self.query,
                                                    self.key)

        mock_npu_rotary.assert_called_once()
        self.assertEqual(result_q.shape, self.query.shape)
        self.assertEqual(result_k.shape, self.key.shape)

        # restore rotary_dim
        self.layer.rotary_dim = org_rotary_dim


class MockRopeModule:

    def __init__(self, max_seq_len=2048, is_neox_style=True):
        self.max_seq_len = max_seq_len
        self.is_neox_style = is_neox_style
        self.cos_cached = None
        self.sin_cached = None
        self.rotary_dim = 1
        self.base = 1


class TestAscendDeepseekScalingRotaryEmbedding(TestBase):

    def setUp(self):
        # Common setup for tests
        self.positions = torch.tensor([1, 2, 3])
        self.query = torch.randn(3, 1, 32, dtype=torch.float16)
        self.key = torch.randn(3, 1, 32, dtype=torch.float16)
        self.head_size = 32
        self.rotary_dim = self.head_size
        self.max_position = 16
        self.rope_theta = 10000
        self.is_neox_style = True
        self.scaling_factor = 1
        self.layer = None

    def _create_layer(self):
        self.layer = DeepseekScalingRotaryEmbedding(
            self.head_size, self.rotary_dim, self.max_position,
            self.rope_theta, self.is_neox_style, self.scaling_factor,
            torch.float16)
        return self.layer

    @patch("vllm.platforms.current_platform.device_type",
           new=torch.device("cpu"))
    @patch("vllm_ascend.ops.rotary_embedding.NPUPlatform",
           new_callable=PropertyMock)
    def test_native_rope_deepseek_forward_base(self, mock_npuplatform):
        mock_npuplatform.device_type = torch.device("cpu")
        self.layer = self._create_layer()
        with patch("vllm_ascend.ops.rotary_embedding._rope_forward_oot",
                   return_value=(self.query,
                                 self.key)) as mock_rope_forward_oot:
            q_pe, k_pe = self.layer.forward(self.positions, self.query,
                                            self.key)
        mock_rope_forward_oot.assert_called_once()
        assert q_pe.shape == self.query.shape
        assert k_pe.shape == self.key.shape

    @patch('vllm_ascend.ops.rotary_embedding._rope_forward_oot')
    @patch("vllm.platforms.current_platform.device_type",
           new=torch.device("cpu"))
    @patch("vllm_ascend.ops.rotary_embedding.NPUPlatform",
           new_callable=PropertyMock)
    def test_native_rope_deepseek_forward_key_reshaping(
            self, mock_npuplatform, mock_rope_forward_oot):
        mock_npuplatform.device_type = torch.device("cpu")
        self.layer = self._create_layer()

        key = torch.randn(1, 32)

        mock_rope_forward_oot.return_value = (self.query, key)

        q_pe, k_pe = self.layer.forward(self.positions, self.query, key)
        mock_rope_forward_oot.assert_called_once()
        assert q_pe.shape == self.query.shape
        assert k_pe.shape == key.shape

    @patch('vllm_ascend.ops.rotary_embedding._rope_forward_oot')
    @patch("vllm.platforms.current_platform.device_type",
           new=torch.device("cpu"))
    @patch("vllm_ascend.ops.rotary_embedding.NPUPlatform",
           new_callable=PropertyMock)
    def test_native_rope_deepseek_forward_non_neox_style(
            self, mock_npuplatform, mock_rope_forward_oot):
        mock_npuplatform.device_type = torch.device("cpu")
        self.layer = self._create_layer()

        mock_rope_forward_oot.return_value = (self.query, self.key)

        q_pe, k_pe = self.layer.forward(self.positions, self.query, self.key)

        mock_rope_forward_oot.assert_called_once()
        assert q_pe.shape == self.query.shape
        assert k_pe.shape == self.key.shape

    @patch("vllm.platforms.current_platform.device_type",
           new=torch.device("cpu"))
    @patch("vllm_ascend.ops.rotary_embedding.NPUPlatform",
           new_callable=PropertyMock)
    def test_basic_case(self, mock_npuplatform):
        # Test with standard values
        mock_npuplatform.device_type = torch.device("cpu")
        self.layer = self._create_layer()
        num_rotations = 100
        dim = 512
        base = 10000
        max_position_embeddings = 2048

        result = self.layer._yarn_find_correction_dim(num_rotations, dim, base,
                                                      max_position_embeddings)

        # Calculate expected value manually
        expected = (dim * torch.log(
            torch.tensor(max_position_embeddings) /
            (num_rotations * 2 * torch.pi))) / (2 *
                                                torch.log(torch.tensor(base)))

        self.assertTrue(torch.allclose(result, expected))

    @patch("vllm.platforms.current_platform.device_type",
           new=torch.device("cpu"))
    @patch("vllm_ascend.ops.rotary_embedding.NPUPlatform",
           new_callable=PropertyMock)
    def test_yarn_get_mscale(self, mock_npuplatform):
        mock_npuplatform.device_type = torch.device("cpu")
        self.layer = self._create_layer()

        # test_scale_less_than_or_equal_1
        self.assertEqual(self.layer._yarn_get_mscale(scale=0.5), 1.0)
        self.assertEqual(self.layer._yarn_get_mscale(scale=1.0), 1.0)
        self.assertEqual(self.layer._yarn_get_mscale(scale=0.999), 1.0)

        # test_scale_greater_than_1:
        test_cases = [(2.0, 1.0, 1.0 + 0.1 * math.log(2.0)),
                      (10.0, 1.0, 1.0 + 0.1 * math.log(10.0)),
                      (5.0, 2.0, 1.0 + 0.2 * math.log(5.0)),
                      (math.e, 1.0, 1.0 + 0.1)]

        for scale, mscale, expected in test_cases:
            result = self.layer._yarn_get_mscale(scale, mscale)
            self.assertAlmostEqual(
                result,
                expected,
                places=6,
                msg=f"Failed for scale={scale}, mscale={mscale}")


class TestAscendMRotaryEmbedding(unittest.TestCase):

    def setUp(self):
        # Common setup for tests
        self.number_tokens = 3
        self.num_head = 8
        self.num_kvhead = 8
        self.head_size = 128
        self.max_position_embeddings = 128000
        self.is_neox_style = True
        self.rope_theta = 1000000.0
        self.positions_1d = torch.tensor([1, 2, 3])
        self.positions_2d = torch.randint(1, 10, (3, self.number_tokens))

        self.query = torch.randn(
            (self.number_tokens, self.num_head * self.head_size),
            dtype=torch.bfloat16)
        self.key = torch.randn(
            (self.number_tokens, self.num_kvhead * self.head_size),
            dtype=torch.bfloat16)

        # Qwen2.5-VL mrope section case
        self.mrope_section = [16, 24, 24]

        self.layer = MRotaryEmbedding(self.head_size,
                                      self.head_size,
                                      self.max_position_embeddings,
                                      base=self.rope_theta,
                                      is_neox_style=self.is_neox_style,
                                      dtype=torch.bfloat16,
                                      mrope_section=self.mrope_section)

        self.mock_config = MagicMock()
        self.mock_config.torchair_graph_config.enabled = False

    def _create_vllm_config(self):
        vllm_config = VllmConfig()
        model_config = ModelConfig(MODEL_VL,
                                   tokenizer=MODEL_VL,
                                   max_model_len=MAX_NUM_BATCHED_TOKEND)
        model_config.hf_config = PretrainedConfig()
        vllm_config.model_config = model_config
        return vllm_config

    @patch('torch_npu.npu_mrope')
    @patch('vllm_ascend.platform.NPUPlatform.get_cpu_architecture')
    @patch('vllm.config.ModelConfig.__post_init__', MagicMock())
    @patch('vllm.config.VllmConfig.__post_init__', MagicMock())
    @patch('vllm.distributed.parallel_state._DP', MagicMock(world_size=1))
    @patch('vllm.distributed.parallel_state._TP', MagicMock(world_size=1))
    def test_forward_oot_1d_positions(self, mock_cpu_arc, mock_npu_mrope):
        mock_cpu_arc.return_value = CpuArchEnum.ARM

        mock_npu_mrope.return_value = (torch.zeros_like(self.query),
                                       torch.zeros_like(self.key))

        vllm_config = self._create_vllm_config()
        with set_ascend_forward_context(None, vllm_config):
            result_q, result_k = self.layer.forward_oot(
                self.positions_1d, self.query, self.key)

        mock_npu_mrope.assert_called_once()
        self.assertFalse(torch.isnan(result_q).any().item())
        self.assertFalse(torch.isnan(result_k).any().item())
        self.assertEqual(result_q.shape, self.query.shape)

    @patch('torch_npu.npu_mrope')
    @patch('vllm_ascend.platform.NPUPlatform.get_cpu_architecture')
    @patch('vllm.config.ModelConfig.__post_init__', MagicMock())
    @patch('vllm.config.VllmConfig.__post_init__', MagicMock())
    @patch('vllm.distributed.parallel_state._DP', MagicMock(world_size=1))
    @patch('vllm.distributed.parallel_state._TP', MagicMock(world_size=1))
    def test_forward_oot_2d_positions(self, mock_cpu_arc, mock_npu_mrope):
        mock_cpu_arc.return_value = CpuArchEnum.ARM

        mock_npu_mrope.return_value = (torch.zeros_like(self.query),
                                       torch.zeros_like(self.key))

        vllm_config = self._create_vllm_config()
        with set_ascend_forward_context(None, vllm_config):
            result_q, result_k = self.layer.forward_oot(
                self.positions_2d, self.query, self.key)

        mock_npu_mrope.assert_called_once()
        self.assertFalse(torch.isnan(result_q).any().item())
        self.assertFalse(torch.isnan(result_k).any().item())
        self.assertEqual(result_q.shape, self.query.shape)



import pytest
import torch
from pytest_mock import MockerFixture
from vllm.model_executor.layers.activation import QuickGELU

from tests.ut.base import PytestBase
from vllm_ascend.models.qwen2_vl import (AscendQwen2VisionAttention,
                                         AscendQwen2VisionBlock)


class TestAscendQwen2VisionAttention(PytestBase):

    def init_attention(
        self,
        mocker,
        embed_dim=1000,
        num_heads=10,
        projection_size=100,
        quant_config=None,
        prefix="",
    ):
        mocker_attn = mocker.patch(
            "vllm_ascend.models.qwen2_vl.Qwen2VisionAttention.__init__")

        attention = AscendQwen2VisionAttention(
            embed_dim=embed_dim,
            num_heads=num_heads,
            projection_size=projection_size,
            quant_config=quant_config,
            prefix=prefix,
        )
        args, kwargs = mocker_attn.call_args
        assert args == (embed_dim, num_heads, projection_size, None, "")
        assert not kwargs
        attention.num_attention_heads_per_partition = num_heads
        return attention

    def test_attn_init_should_normal(self, mocker: MockerFixture):
        embed_dim = 1000
        num_heads = 10
        projection_size = 100
        quant_config = None
        prefix = ""
        vit = self.init_attention(
            embed_dim=embed_dim,
            num_heads=num_heads,
            projection_size=projection_size,
            quant_config=quant_config,
            prefix=prefix,
            mocker=mocker,
        )
        assert vit.hidden_size_per_attention_head == 10

    def test_attn_init_should_raise_error(self, mocker: MockerFixture):
        embed_dim = 1000
        num_heads = 7
        projection_size = 100
        quant_config = None
        prefix = ""
        with pytest.raises(AssertionError):
            # projection_size should divided by num heads
            self.init_attention(
                mocker=mocker,
                embed_dim=embed_dim,
                num_heads=num_heads,
                projection_size=projection_size,
                quant_config=quant_config,
                prefix=prefix,
            )

    def test_attn_forward(self, mocker: MockerFixture):
        attention = self.init_attention(mocker=mocker)
        mocker.patch("torch.nn.Module.__setattr__")
        mocker.patch("torch.nn.Module.__getattr__")
        mocker.patch("torch.nn.Module.__delattr__")
        x = torch.rand((100, 3, 10 * 3 * 128))  # s,b, head*3*head_dim
        cu_seqlens = torch.tensor([10, 50, 100])
        cos = torch.rand((1, 100, 1, 128))
        sin = torch.rand((1, 100, 1, 128))

        qkv = lambda x: (x, 0)  # noqa
        split_qkv = lambda x: [  #noqa
            torch.rand((100, 3, 10, 128)) for i in range(3)
        ]  # noqa
        npu_rotary_mul = lambda q, cos, sin: q  # noqa
        _npu_flash_attention_unpad = lambda **kwargs: kwargs["out"]  # noqa
        proj = lambda x: (x, 0)  # noqa

        mocker_qkv = mocker.patch.object(attention, "qkv", side_effect=qkv)
        mocker_split_qkv = mocker.patch.object(
            attention,
            "split_qkv",
            side_effect=split_qkv,
        )
        mocker_npu_rotary_mul = mocker.patch("torch_npu.npu_rotary_mul",
                                             side_effect=npu_rotary_mul)
        mocker_npu_flash_attention_unpad = mocker.patch(
            "torch_npu._npu_flash_attention_unpad",
            side_effect=_npu_flash_attention_unpad,
        )
        mocker_proj = mocker.patch.object(attention, "proj", side_effect=proj)
        attention.__dict__["qkv"] = mocker_qkv
        attention.__dict__["split_qkv"] = mocker_split_qkv
        attention.__dict__["npu_rotary_mul"] = mocker_npu_rotary_mul
        attention.__dict__["_npu_flash_attention_unpad"] = (
            mocker_npu_flash_attention_unpad)
        attention.__dict__["proj"] = mocker_proj

        output = attention.forward(
            x=x,
            cu_seqlens=cu_seqlens,
            cos=cos,
            sin=sin,
        )
        qkv_args, qkv_kwargs = mocker_qkv.call_args
        assert qkv_args == (x, )
        assert not qkv_kwargs

        split_qkv_args, split_qkv_kwargs = mocker_split_qkv.call_args
        assert split_qkv_args == (x, )
        assert not split_qkv_kwargs

        npu_rotary_mul_args, npu_rotary_mul_kwargs = mocker_npu_rotary_mul.call_args
        assert npu_rotary_mul_args[1:] == (cos, sin)
        assert npu_rotary_mul_args[0].shape == torch.Size([3, 100, 10, 128])
        assert not npu_rotary_mul_kwargs

        assert output.shape == torch.Size([100, 3, 1280])


class TestAscendQwen2VisionBlock(PytestBase):

    def init_vision_block(
        self,
        mocker,
        dim=100,
        num_heads=10,
        mlp_ratio=0.5,
    ):
        mocker_vit = mocker.patch(
            "vllm.model_executor.models.qwen2_vl.Qwen2VisionBlock.__init__",
            return_value=None,
        )

        mocker_attn = mocker.patch(
            "vllm_ascend.models.qwen2_vl.AscendQwen2VisionAttention.__init__",
            return_value=None,
        )

        mocker.patch("torch.nn.Module.__setattr__")
        mocker.patch("torch.nn.Module.__getattr__")
        mocker.patch("torch.nn.Module.__delattr__")
        vision_block = AscendQwen2VisionBlock(
            dim=dim,
            num_heads=num_heads,
            mlp_ratio=mlp_ratio,
        )
        args, kwargs = mocker_vit.call_args
        assert args == (dim, num_heads, mlp_ratio, QuickGELU, None, None, "")
        assert not kwargs

        args1, kwargs1 = mocker_attn.call_args
        assert not args1
        assert kwargs1 == {
            "embed_dim": dim,
            "num_heads": num_heads,
            "projection_size": dim,
            "quant_config": None,
            "prefix": ".attn",
        }
        return vision_block

    def test_init_vision_block_should_normal(
        self,
        mocker: MockerFixture,
    ):
        vision_block = self.init_vision_block(mocker)
        assert isinstance(vision_block, AscendQwen2VisionBlock)

    def test_vision_block_forward(self, mocker: MockerFixture):
        x = torch.randint(1, 100, (100, 3, 1280))  # s,b,d
        cu_seqlens = torch.tensor([10, 50, 100])
        cos = torch.rand((1, 100, 1, 128))
        sin = torch.rand((1, 100, 1, 128))
        vision_block = self.init_vision_block(mocker)
        mocker_attn = mocker.patch.object(vision_block, "attn", return_value=x)
        mocker_mlp = mocker.patch.object(vision_block, "mlp", return_value=x)
        vision_block.__dict__["attn"] = mocker_attn
        vision_block.__dict__["mlp"] = mocker_mlp

        output = vision_block.forward(x.clone(), cu_seqlens, cos, sin)

        _, attn_kwargs = mocker_attn.call_args
        assert attn_kwargs == {
            "cu_seqlens": cu_seqlens,
            "cos": cos,
            "sin": sin,
        }

        assert torch.all(x * 3 == output)



#
# Copyright (c) 2025 Huawei Technologies Co., Ltd. All Rights Reserved.
# This file is a part of the vllm-ascend project.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

import vllm_ascend.patch.platform.patch_config  # noqa
import vllm_ascend.patch.platform.patch_distributed  # noqa
import vllm_ascend.patch.platform.patch_mamba_config  # noqa
import vllm_ascend.patch.platform.patch_sched_yield  # noqa

if os.getenv("DYNAMIC_EPLB", "false") == "true" or os.getenv(
        "EXPERT_MAP_RECORD", "false") == "true":
    import vllm_ascend.patch.platform.patch_multiproc_executor  # noqa

if os.getenv("SHM_BARRIER", "true") == "true":
    import vllm_ascend.patch.platform.patch_core  # noqa
    import vllm_ascend.patch.platform.patch_message_queue  # noqa



#
# Copyright (c) 2025 Huawei Technologies Co., Ltd. All Rights Reserved.
# Copyright 2023 The vLLM team.
#
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Adapted from vllm/model_executor/models/qwen2_vl.py
# This file is a part of the vllm-ascend project.

import torch
import vllm.envs as envs_vllm
from vllm.config import ParallelConfig

from vllm_ascend.utils import is_310p


def parallel_config_get_dp_port(self) -> int:
    """
    We might need to initialize process groups in multiple
    processes that is related to data parallelism,
    e.g. both in the worker and in the engine, which
    can live in different processes. To avoid port conflicts, we
    increment the port number each time we need to initialize a
    new process group related to data parallelism.
    """
    answer = self.data_parallel_master_port
    self.data_parallel_master_port += 1

    # NOTE: Get port from envs directly when using torchrun
    port = envs_vllm.VLLM_DP_MASTER_PORT if envs_vllm.VLLM_DP_MASTER_PORT else answer
    return port


ParallelConfig.get_next_dp_init_port = parallel_config_get_dp_port


class NullHandle:

    def __init__(self):
        pass

    def wait(self):
        pass


def communication_adaptation_310p():

    def broadcast310p_wrapper(fn):

        def broadcast310p(tensor, src, group=None, async_op=False):
            if tensor.device == torch.device('cpu'):
                return fn(tensor, src, group, async_op)
            rank = torch.distributed.get_rank(group)
            world_size = torch.distributed.get_world_size(group)
            tensor_list = [torch.empty_like(tensor) for _ in range(world_size)]
            tensor_list[rank] = tensor
            torch.distributed.all_gather(tensor_list, tensor, group=group)
            tensor[...] = tensor_list[src]
            if async_op:
                return NullHandle()
            else:
                return None

        return broadcast310p

    torch.distributed.broadcast = broadcast310p_wrapper(
        torch.distributed.broadcast)
    torch.distributed.distributed_c10d.broadcast = broadcast310p_wrapper(
        torch.distributed.distributed_c10d.broadcast)

    def all_reduce_wrapper_310p(fn):

        def all_reduce(
            tensor,
            op=torch.distributed.ReduceOp.SUM,
            group=None,
            async_op=False,
        ):
            if tensor.dtype != torch.int64:
                return fn(tensor, op, group, async_op)
            rank = torch.distributed.get_rank(group)
            world_size = torch.distributed.get_world_size(group)
            tensor_list = [torch.empty_like(tensor) for _ in range(world_size)]
            tensor_list[rank] = tensor
            torch.distributed.all_gather(tensor_list, tensor, group=group)
            if op == torch.distributed.ReduceOp.SUM:
                return torch.stack(tensor_list).sum(0)
            elif op == torch.distributed.ReduceOp.MAX:
                return torch.tensor(
                    torch.stack(tensor_list).cpu().numpy().max(0),
                    device=tensor.device,
                )
            else:
                raise RuntimeError(f"not implement op {op}")

        return all_reduce

    torch.distributed.all_reduce = all_reduce_wrapper_310p(
        torch.distributed.all_reduce)
    torch.distributed.distributed_c10d.all_reduce = all_reduce_wrapper_310p(
        torch.distributed.distributed_c10d.all_reduce)


if is_310p():
    communication_adaptation_310p()



import gc
from typing import Tuple

import pytest
import torch
import torch_npu  # noqa: F401

import vllm_ascend.platform  # noqa: F401
from vllm_ascend.utils import enable_custom_op

enable_custom_op()

# Test parameters
DTYPES = [torch.int32]
#SHAPES = [(100,), (5, 20), (3, 4, 5)]  # Various tensor shapes
#SHAPES = [(3, 4, 8), (3, 4, 5)]  # Various tensor shapes
SHAPES = [(3, 4, 3)]
DEVICES = [f"npu:{0}"]
SEEDS = [0]


def get_masked_input_and_mask_ref(
        input_: torch.Tensor, org_vocab_start_index: int,
        org_vocab_end_index: int, num_org_vocab_padding: int,
        added_vocab_start_index: int,
        added_vocab_end_index: int) -> Tuple[torch.Tensor, torch.Tensor]:
    """Reference implementation for verification"""
    org_vocab_mask = (input_ >= org_vocab_start_index) & (
        input_ < org_vocab_end_index)
    added_vocab_mask = (input_ >= added_vocab_start_index) & (
        input_ < added_vocab_end_index)
    added_offset = added_vocab_start_index - (
        org_vocab_end_index - org_vocab_start_index) - num_org_vocab_padding
    valid_offset = (org_vocab_start_index *
                    org_vocab_mask) + (added_offset * added_vocab_mask)
    vocab_mask = org_vocab_mask | added_vocab_mask
    masked_input = vocab_mask * (input_ - valid_offset)
    return masked_input, ~vocab_mask


@pytest.mark.parametrize("shape", SHAPES)
@pytest.mark.parametrize("dtype", DTYPES)
@pytest.mark.parametrize("device", DEVICES)
@pytest.mark.parametrize("seed", SEEDS)
@torch.inference_mode()
def test_get_masked_input_and_mask(
    shape: Tuple[int, ...],
    dtype: torch.dtype,
    device: str,
    seed: int,
) -> None:
    # Set random seed
    torch.manual_seed(seed)
    torch.set_default_device(device)

    # Generate random input tensor
    input_tensor = torch.randint(0, 1000, shape, dtype=dtype)

    # Test parameters
    test_case = {
        "org_start": 100,
        "org_end": 200,
        "padding": 0,
        "added_start": 300,
        "added_end": 400,
    }

    # Get reference result
    ref_masked_input, ref_mask = get_masked_input_and_mask_ref(
        input_tensor, test_case["org_start"], test_case["org_end"],
        test_case["padding"], test_case["added_start"], test_case["added_end"])

    # Get custom op result
    print("input_tensor:", input_tensor)
    custom_masked_input, custom_mask = torch.ops._C_ascend.get_masked_input_and_mask(
        input_tensor, test_case["org_start"], test_case["org_end"],
        test_case["padding"], test_case["added_start"], test_case["added_end"])

    ref_masked_input = ref_masked_input.to(dtype)
    print("custom_masked_input:", custom_masked_input)
    print("ref_masked_input:", ref_masked_input)
    print("custom_mask:", custom_mask)
    print("ref_mask:", ref_mask)
    # Compare results
    torch.testing.assert_close(
        custom_masked_input,
        ref_masked_input,
        rtol=1e-5,
        atol=1e-5,
        msg=f"Masked input mismatch for case: {test_case}")
    torch.testing.assert_close(custom_mask,
                               ref_mask,
                               rtol=1e-5,
                               atol=1e-5,
                               msg=f"Mask mismatch for case: {test_case}")
    gc.collect()
    torch.npu.empty_cache()
    torch.npu.reset_peak_memory_stats()



import types

import torch
import torch.nn as nn
import torchair
from torchair import patch_for_hcom
from vllm.attention.layer import Attention
from vllm.config import (CUDAGraphMode, VllmConfig,
                         get_layers_from_vllm_config, set_current_vllm_config)
from vllm.forward_context import BatchDescriptor, get_forward_context
from vllm.model_executor.model_loader import get_model_loader
from vllm.model_executor.model_loader.utils import (
    process_weights_after_loading, set_default_torch_dtype)
from vllm.v1.core.sched.output import SchedulerOutput
from vllm.v1.sample.metadata import SamplingMetadata
from vllm.v1.spec_decode.metadata import SpecDecodeMetadata

from vllm_ascend.ascend_config import get_ascend_config
from vllm_ascend.ascend_forward_context import set_ascend_forward_context
from vllm_ascend.attention.utils import AscendCommonAttentionMetadata
from vllm_ascend.patch.worker.patch_deepseek_mtp import \
    AscendDeepSeekMTP as DeepSeekMTP
from vllm_ascend.spec_decode.interface import Proposer, SpecDcodeType
from vllm_ascend.torchair.models.torchair_deepseek_mtp import \
    TorchairDeepSeekMTP
from vllm_ascend.torchair.utils import (TORCHAIR_CACHE_DIR,
                                        TorchairCommonAttentionMetadata)
from vllm_ascend.utils import ProfileExecuteDuration, lmhead_tp_enable

PADDING_SLOT_ID = -1


class MtpProposer(Proposer):

    def __init__(
        self,
        vllm_config: VllmConfig,
        device,
        runner,
    ):
        self.name = SpecDcodeType.MTP
        self.vllm_config = vllm_config
        self.device = device
        self.runner = runner
        self.num_speculative_tokens = vllm_config.speculative_config.num_speculative_tokens

        # persistent buffers for graph
        self.input_ids = torch.zeros(self.runner.max_num_tokens,
                                     dtype=torch.int32,
                                     device=self.device)
        self.positions = torch.zeros(self.runner.max_num_tokens,
                                     dtype=torch.int64,
                                     device=self.device)
        self.hidden_states = torch.zeros(
            (self.runner.max_num_tokens,
             vllm_config.model_config.get_hidden_size()),
            dtype=self.runner.dtype,
            device=self.device)
        self.torchair_compiled_model = None  # type: ignore
        self.torchair_compiled_models = {}  # type: ignore
        self.torchair_graph_enabled = get_ascend_config(
        ).torchair_graph_config.enabled
        self.enable_shared_expert_dp = get_ascend_config(
        ).enable_shared_expert_dp
        # We need +1 here because the arange is used to set query_start_loc,
        # which has one more element than batch_size.
        self.arange = torch.arange(vllm_config.scheduler_config.max_num_seqs +
                                   1,
                                   device=self.runner.device,
                                   dtype=torch.int32)
        self.use_sparse = hasattr(vllm_config.model_config.hf_config,
                                  "index_topk")

    def load_model(self, model) -> None:
        loader = get_model_loader(self.vllm_config.load_config)

        target_attn_layer_names = set(
            get_layers_from_vllm_config(self.vllm_config, Attention).keys())
        draft_model_config = \
            self.vllm_config.speculative_config.draft_model_config
        target_device = self.vllm_config.device_config.device

        with set_default_torch_dtype(
                draft_model_config.dtype), set_current_vllm_config(
                    self.vllm_config):
            if self.torchair_graph_enabled or (
                    self.enable_shared_expert_dp
                    and self.vllm_config.model_config.use_mla):
                self.model = TorchairDeepSeekMTP(
                    vllm_config=self.vllm_config).to(target_device)
            else:
                self.model = DeepSeekMTP(
                    vllm_config=self.vllm_config).to(target_device)

        draft_attn_layer_names = (
            get_layers_from_vllm_config(self.vllm_config, Attention).keys() -
            target_attn_layer_names)

        assert len(draft_attn_layer_names) == 1
        self.attn_layer_name = list(draft_attn_layer_names)

        self.model.load_weights(
            loader.get_all_weights(
                self.vllm_config.speculative_config.draft_model_config,
                self.model))
        process_weights_after_loading(self.model, draft_model_config,
                                      target_device)

    @torch.inference_mode()
    def dummy_run(self,
                  num_tokens: int,
                  with_prefill: bool = False,
                  skip_attn: bool = False,
                  num_reqs: int = 0,
                  num_tokens_across_dp=None,
                  aclgraph_runtime_mode: CUDAGraphMode = CUDAGraphMode.NONE,
                  batch_descriptor=None,
                  dummy_compute_logits=lambda hidden_states: None) -> None:
        if not self.torchair_graph_enabled:
            # TODO: adapt enable_dbo later
            (num_tokens, num_tokens_across_dp, with_prefill,
             _) = self.runner._sync_metadata_across_dp(num_tokens,
                                                       with_prefill, False)

        moe_comm_type = self.runner._select_moe_comm_method(
            num_tokens, with_prefill)

        is_running_torchair = self.torchair_graph_enabled and \
            not with_prefill

        if is_running_torchair:
            skip_attn = False
        if skip_attn:
            attn_metadata = None
        else:
            common_attn_metadata = TorchairCommonAttentionMetadata(
                num_reqs=num_reqs,
                num_actual_tokens=1,
                actual_seq_lengths_q=self.runner.actual_seq_lengths_q,
                attn_mask=self.runner.attn_mask,
                spec_attn_mask=self.runner.spec_attn_mask,
                decode_token_per_req=self.runner.decode_token_per_req,
            )
            attn_metadata = self.runner.attn_metadata_builder.build_torchair_graph_dummy(
                common_attn_metadata)

        input_ids = self.input_ids[:num_tokens]
        positions = self.positions[:num_tokens]
        previous_hidden_states = self.hidden_states[:num_tokens]
        for _ in range(self.num_speculative_tokens):
            with set_ascend_forward_context(
                    attn_metadata,
                    self.vllm_config,
                    num_tokens=num_tokens,
                    with_prefill=with_prefill,
                    num_tokens_across_dp=num_tokens_across_dp,
                    reserved_mc2_mask=self.runner.reserved_mc2_mask,
                    moe_comm_type=moe_comm_type,
                    in_profile_run=self.runner.in_profile_run,
                    num_actual_tokens=0,
                    aclgraph_runtime_mode=aclgraph_runtime_mode,
                    batch_descriptor=batch_descriptor):
                if is_running_torchair:
                    assert attn_metadata is not None
                    torch._dynamo.mark_static(input_ids)
                    torch._dynamo.mark_static(positions)
                    torch._dynamo.mark_static(previous_hidden_states)
                    torch._dynamo.mark_static(attn_metadata.decode.block_table)
                    torch._dynamo.mark_static(
                        attn_metadata.decode.input_positions)
                    if hasattr(attn_metadata.decode, "sin"):
                        torch._dynamo.mark_static(attn_metadata.decode.sin)
                        torch._dynamo.mark_static(attn_metadata.decode.cos)
                    torch._dynamo.mark_static(get_forward_context().mc2_mask)
                    torch._dynamo.mark_static(attn_metadata.slot_mapping)
                    torch._dynamo.mark_static(attn_metadata.decode.attn_mask)
                    torchair_compiled_model = self._get_torchair_lazy_compiled_model(
                        num_tokens)
                    torchair_compiled_model(
                        input_ids=input_ids,
                        positions=positions,
                        hidden_states=previous_hidden_states,
                        inputs_embeds=None,
                        intermediate_tensors=None,
                        attn_metadata=attn_metadata,
                        kv_caches=self.runner.kv_caches[-1:],
                        spec_step_idx=0)
                else:
                    self.model(input_ids=input_ids,
                               positions=positions,
                               hidden_states=previous_hidden_states)
                dummy_compute_logits(previous_hidden_states)
            if with_prefill:
                break

    def generate_token_ids(self,
                           valid_sampled_token_ids: list[list[int]],
                           sampling_metadata: SamplingMetadata = None,
                           scheduler_output: SchedulerOutput = None,
                           spec_decode_metadata: SpecDecodeMetadata = None,
                           positions: torch.Tensor = None,
                           num_scheduled_tokens: int = 0,
                           hidden_states: torch.Tensor = None,
                           attn_metadata=None,
                           aux_hidden_states: torch.Tensor = None):
        if attn_metadata is not None and isinstance(attn_metadata, dict):
            attn_metadata = attn_metadata['model.layers.0.self_attn.attn']
        next_token_ids: list[int] = []
        for i, token_ids in enumerate(valid_sampled_token_ids):
            if token_ids:
                # Common case.
                next_token_id = token_ids[-1]
            else:
                # Partial prefill (rare case).
                # Get the next token id from the request state.
                req_id = self.runner.input_batch.req_ids[i]
                req_state = self.runner.requests[req_id]
                seq_len = (req_state.num_computed_tokens +
                           scheduler_output.num_scheduled_tokens[req_id])
                next_token_id = req_state.get_token_id(seq_len)
            next_token_ids.append(next_token_id)
        next_token_ids = torch.tensor(next_token_ids,
                                      dtype=torch.int32,
                                      device=self.device)
        accepted_token_indices = None
        if spec_decode_metadata is None:
            # input_ids can be None for multimodal models.
            target_token_ids = self.runner.input_ids[:num_scheduled_tokens]
            target_positions = positions[:num_scheduled_tokens]
            target_hidden_states = hidden_states[:num_scheduled_tokens]
            target_slot_mapping = attn_metadata.slot_mapping
            cu_num_tokens = attn_metadata.query_start_loc
        else:
            # TODO(woosuk): Refactor this.
            num_draft_tokens = spec_decode_metadata.num_draft_tokens
            num_rejected_tokens = [
                n + 1 - len(valid_sampled_token_ids[i]) if n > 0 else 0
                for i, n in enumerate(num_draft_tokens)
            ]
            num_rejected_tokens = torch.tensor(
                num_rejected_tokens,
                dtype=torch.int32,
                device=self.device,
            )
            cu_num_tokens, accepted_token_indices, target_token_ids, \
                target_positions, target_hidden_states, target_slot_mapping = self._prepare_inputs(
                attn_metadata.query_start_loc,
                num_rejected_tokens,
                self.runner.input_ids[:num_scheduled_tokens],
                positions[:num_scheduled_tokens],
                hidden_states[:num_scheduled_tokens],
                attn_metadata.slot_mapping[:num_scheduled_tokens],
                is_torchair_graph=self.runner._build_drafter_prepare_inputs_torchair_param(),
            )

        draft_token_ids = self._propose(
            target_token_ids=target_token_ids,
            target_positions=target_positions,
            target_hidden_states=target_hidden_states,
            target_slot_mapping=target_slot_mapping,
            next_token_ids=next_token_ids,
            cu_num_tokens=cu_num_tokens,
            block_table=attn_metadata.block_tables,
            sampling_metadata=sampling_metadata,
            token_indices=accepted_token_indices)
        spec_token_ids = draft_token_ids.tolist()
        return spec_token_ids

    def _prepare_inputs(
        self,
        # [batch_size + 1]
        cu_target_query_lens: torch.Tensor,
        # [batch_size]
        num_rejected_tokens: torch.Tensor,
        token_ids: torch.Tensor,
        positions: torch.Tensor,
        hidden_states: torch.Tensor,
        slot_mapping: torch.Tensor,
        is_torchair_graph: bool = False
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor,
               torch.Tensor, torch.Tensor]:
        # cu_target_query_lens: [0, a, a + b, a + b + c]
        # num_rejected_tokens: [n1, n2, n3]
        # num_tokens_per_req: [a - n1, b - n2, c - n3]
        # cu_num_tokens: [0, a - n1, a + b - n1 - n2, a + b + c - n1 - n2 - n3]
        # token_indices: [0, 1, ..., a - n1 - 1,
        #                 a, a + 1, ..., a + b - n2 - 1,
        #                 a + b, a + b + 1, ..., a + b + c - n3 - 1]
        # [0, a, a + b, a + b + c] -> [a, b, c]
        query_len_per_req = (cu_target_query_lens[1:] -
                             cu_target_query_lens[:-1])
        # [a, b, c] -> [a - n1, b - n2, c - n3]
        num_tokens_per_req = query_len_per_req - num_rejected_tokens
        if is_torchair_graph:
            cu_num_tokens = cu_target_query_lens
            relative_index = query_len_per_req - num_rejected_tokens - 1
            token_indices = cu_num_tokens[:-1] + relative_index
            # the seq len of each bath is padded to 1+num_speculative_tokens, thus input is same as the main model
            target_token_ids = token_ids
            target_positions = positions
            target_hidden_states = hidden_states
            target_slot_mapping = slot_mapping
        else:
            cu_num_tokens = torch.empty_like(cu_target_query_lens)
            torch.cumsum(num_tokens_per_req, dim=0, out=cu_num_tokens[1:])
            cu_num_tokens[0] = 0

            # FIXME(woosuk): Avoid synchronization.
            num_tokens = cu_num_tokens[-1].item()
            token_indices = torch.zeros(
                num_tokens,
                dtype=torch.int32,
                device=cu_num_tokens.device,
            )

            BLOCK_SIZE = 1024
            self._prepare_input_kernel(
                token_indices,
                cu_target_query_lens,
                cu_num_tokens,
                block_size=BLOCK_SIZE,
            )
            target_token_ids = token_ids[token_indices]
            target_positions = positions[token_indices]
            target_hidden_states = hidden_states[token_indices]
            target_slot_mapping = slot_mapping[token_indices]
        return cu_num_tokens, token_indices, target_token_ids, target_positions, target_hidden_states, target_slot_mapping

    def _propose(
            self,
            # [num_tokens]
            target_token_ids: torch.Tensor,
            # [num_tokens]
            target_positions: torch.Tensor,
            # [num_tokens, hidden_size]
            target_hidden_states: torch.Tensor,
            # [num_tokens]
            target_slot_mapping: torch.Tensor,
            # [batch_size]
            next_token_ids: torch.Tensor,
            # [batch_size + 1] starting with 0
            cu_num_tokens: torch.Tensor,
            # [batch_size, max_num_blocks_per_req]
            block_table: torch.Tensor,
            sampling_metadata: SamplingMetadata,
            token_indices=None) -> torch.Tensor:
        num_tokens = target_token_ids.shape[0]
        batch_size = next_token_ids.shape[0]
        last_token_indices = cu_num_tokens[1:] - 1

        # Shift the input ids by one token.
        # E.g., [a1, b1, b2, c1, c2, c3] -> [b1, b2, c1, c2, c3, c3]
        self.input_ids[:num_tokens - 1] = target_token_ids[1:]
        # Replace the last token with the next token.
        # E.g., [b1, b2, c1, c2, c3, c3] -> [a2, b2, b3, c2, c3, c4]
        if token_indices is not None and self.torchair_graph_enabled:
            last_token_indices = token_indices

        self.input_ids[last_token_indices] = next_token_ids

        query_lens = cu_num_tokens[1:] - cu_num_tokens[:-1]
        max_query_len = query_lens.max().item()

        # FIXME: reorder_batch() needs to be called before build()
        # because fields of attn_metadata_builder needs to be updated.
        # However, currently reorder_batch() takes input_batch and
        # scheduler_output as arguments, we should probably refactor
        # the method to use new data structures which are independent
        # from input_batch and scheduler_output.
        # self.runner.attn_metadata_builder.reorder_batch(
        #     input_batch=self.runner.input_batch,
        #     scheduler_output=self.runner.scheduler_output,
        # )
        is_running_torchair = self.torchair_graph_enabled and \
            not self.runner.with_prefill

        if is_running_torchair:
            # Torchair graph mode, padding is same as the main model
            num_input_tokens = self.runner.graph_pad_size
        elif (self.runner.use_aclgraph
              and num_tokens <= self.runner.aclgraph_batch_sizes[-1]):
            # Acl graph mode, add padding to the batch size
            num_input_tokens = self.vllm_config.pad_for_cudagraph(num_tokens)
        else:
            # Eager mode, no padding needed
            num_input_tokens = num_tokens

        seq_lens = target_positions[last_token_indices] + 1
        seq_lens = seq_lens.int()
        common_attn_metadata = AscendCommonAttentionMetadata(
            query_start_loc=cu_num_tokens[:batch_size + 1],
            query_start_loc_cpu=cu_num_tokens[:batch_size + 1].cpu(),
            seq_lens_cpu=seq_lens.cpu(),
            num_reqs=batch_size,
            num_actual_tokens=num_tokens,
            max_query_len=max_query_len,
            actual_seq_lengths_q=self.runner.actual_seq_lengths_q,
            block_table_tensor=self.runner.input_batch.block_table[0].
            get_device_tensor(),
            slot_mapping=target_slot_mapping,
            positions=target_positions,
            attn_mask=self.runner.attn_mask,
            spec_attn_mask=self.runner.spec_attn_mask,
            attn_state=self.runner.attn_state,
            graph_pad_size=self.runner.graph_pad_size,
            decode_token_per_req=self.runner.decode_token_per_req,
            num_computed_tokens_cpu=None,
            seq_lens=None)

        if not self.torchair_graph_enabled:
            builder = self.runner.attn_groups[0][0].get_metadata_builder()
            attn_metadata_mtp = builder.build(0, common_attn_metadata,
                                              self.runner.get_model())

            attn_metadata = {}
            for layer_name in self.attn_layer_name:
                attn_metadata[layer_name] = attn_metadata_mtp

        else:
            attn_metadata = self.runner.attn_metadata_builder.build(
                0, common_attn_metadata, self.runner.get_model())

        self.positions[:num_tokens] = target_positions
        self.hidden_states[:num_tokens] = target_hidden_states

        if not self.torchair_graph_enabled:
            # torch mode need to update num_tokens_across_dp
            # TODO: adapt enable_dbo later
            (num_input_tokens, num_tokens_across_dp, with_prefill,
             _) = self.runner._sync_metadata_across_dp(
                 num_input_tokens, self.runner.with_prefill, False)
        else:
            # torchair mode can reuse self.runner.num_tokens_across_dp
            num_tokens_across_dp = self.runner.num_tokens_across_dp
            with_prefill = self.runner.with_prefill

        moe_comm_type = self.runner._select_moe_comm_method(
            num_input_tokens, with_prefill)
        batch_descriptor = BatchDescriptor(num_tokens=num_input_tokens,
                                           uniform_decode=False)
        aclgraph_runtime_mode, batch_descriptor = \
            self.runner.aclgraph_dispatcher.dispatch(batch_descriptor)

        for step in range(self.num_speculative_tokens):
            with set_ascend_forward_context(
                    attn_metadata,
                    self.vllm_config,
                    num_tokens=num_input_tokens,
                    with_prefill=with_prefill,
                    num_tokens_across_dp=num_tokens_across_dp,
                    reserved_mc2_mask=self.runner.reserved_mc2_mask,
                    moe_comm_type=moe_comm_type,
                    aclgraph_runtime_mode=aclgraph_runtime_mode,
                    batch_descriptor=batch_descriptor,
                    in_profile_run=self.runner.in_profile_run,
                    num_actual_tokens=num_tokens):
                with ProfileExecuteDuration().capture_async('mtp_forward'):
                    model_kwargs = {}
                    model_kwargs["attn_metadata"] = attn_metadata
                    if self.torchair_graph_enabled:
                        model_kwargs["kv_caches"] = self.runner.kv_caches[-1:]
                    if is_running_torchair:
                        torchair_compiled_model = self._get_torchair_lazy_compiled_model(
                            num_input_tokens)
                        hidden_states = torchair_compiled_model(
                            input_ids=self.input_ids[:num_input_tokens],
                            positions=self.positions[:num_input_tokens],
                            hidden_states=self.
                            hidden_states[:num_input_tokens],
                            inputs_embeds=None,
                            intermediate_tensors=None,
                            spec_step_idx=0,
                            **model_kwargs)
                    else:
                        hidden_states = self.model(
                            input_ids=self.input_ids[:num_input_tokens],
                            positions=self.positions[:num_input_tokens],
                            hidden_states=self.hidden_states[:num_input_tokens]
                        )

            num_indices = last_token_indices.shape[0]
            if lmhead_tp_enable():
                if not self.runner.with_prefill:
                    max_num_reqs_across_dp = num_input_tokens
                else:
                    max_num_reqs_across_dp = self.vllm_config.scheduler_config.max_num_seqs
                last_token_indices = nn.functional.pad(
                    last_token_indices,
                    (0, max_num_reqs_across_dp - num_indices))

            sample_hidden_states = hidden_states[last_token_indices]
            logits = self.model.compute_logits(sample_hidden_states)
            if lmhead_tp_enable() and num_indices < logits.shape[0]:
                logits = logits[:num_indices]
                last_token_indices = last_token_indices[:num_indices]
            draft_token_ids = logits.argmax(dim=-1)

            if self.num_speculative_tokens == 1:
                # [batch_size, 1]
                return draft_token_ids.view(-1, 1)

            if step == 0:
                draft_token_ids_list = [draft_token_ids]
            else:
                draft_token_ids_list.append(draft_token_ids)

            # prepare next mtp inputs
            # mtp>1: prefill skip or decode skip last loop
            if with_prefill:
                for _ in range(self.num_speculative_tokens - 1):
                    draft_token_ids_list.append(draft_token_ids)
            if step == self.num_speculative_tokens - 1 or with_prefill:
                break

            if not self.torchair_graph_enabled:
                attn_metadata_i = attn_metadata[self.attn_layer_name[0]]
            else:
                attn_metadata_i = attn_metadata

            if step == 0:
                positions = target_positions[last_token_indices]
                hidden_states = hidden_states[last_token_indices]
                slot_mapping = attn_metadata_i.slot_mapping[last_token_indices]
                attn_metadata_i.slot_mapping.fill_(-1)
                attn_metadata_i.query_start_loc = self.arange[:batch_size + 1]
                last_token_indices = self.arange[:batch_size]
                if attn_metadata_i.num_decode_tokens != 0:
                    attn_metadata_i.num_decode_tokens = batch_size
                if is_running_torchair:
                    attn_metadata_i.num_actual_tokens = batch_size
                    attn_metadata_i.query_lens = [1] * batch_size

            input_ids = draft_token_ids_list[-1].int()
            positions += 1

            if not self.torchair_graph_enabled:
                attn_metadata_i.decode.actual_seq_lengths_q = attn_metadata_i.query_start_loc[
                    1:batch_size + 1].tolist()
                attn_metadata_i.decode.cos = builder.cos_cache[
                    positions].unsqueeze(1).unsqueeze(2)
                attn_metadata_i.decode.sin = builder.sin_cache[
                    positions].unsqueeze(1).unsqueeze(2)

            # NOTE(woosuk): We should handle the case where the draft model
            # generates tokens beyond the max model length. Since it is complex
            # to remove such requests from the batch, we keep them in the batch
            # but adjust the position ids and slot mappings to avoid the
            # out-of-range access during the model execution. The draft tokens
            # generated with this adjustment should be ignored.
            exceeds_max_model_len = positions >= self.runner.model_config.max_model_len
            # Mask out the position ids that exceed the max model length.
            # Otherwise, we may get out-of-range error in RoPE.
            clamped_positions = torch.where(exceeds_max_model_len, 0,
                                            positions)
            # Increment the sequence lengths.
            attn_metadata_i.seq_lens[:batch_size] += 1
            # For the requests that exceed the max model length, we set the
            # sequence length to 1 to minimize their overheads in attention.
            exceeds_max_model_len_cpu = exceeds_max_model_len.to(
                attn_metadata_i.seq_lens.device, non_blocking=False)
            attn_metadata_i.seq_lens[:batch_size].masked_fill_(
                exceeds_max_model_len_cpu, 1)
            # Mask out the slot mappings that exceed the max model length.
            # Otherwise, the KV cache will be inadvertently updated with the
            # padding tokens.
            slot_mapping += 1
            slot_mapping.masked_fill_(exceeds_max_model_len, PADDING_SLOT_ID)

            # copy inputs to buffer for cudagraph
            self.input_ids[:batch_size] = input_ids
            self.positions[:batch_size] = clamped_positions
            self.hidden_states[:hidden_states.shape[0]] = hidden_states
            attn_metadata_i.slot_mapping[:batch_size] = slot_mapping

            if attn_metadata_i.prefill is not None:
                attn_metadata_i.prefill.seq_lens = attn_metadata_i.seq_lens
                attn_metadata_i.prefill.seq_lens_list = attn_metadata_i.prefill.seq_lens.tolist(
                )
                attn_metadata_i.prefill.context_lens = attn_metadata_i.seq_lens
                attn_metadata_i.prefill.input_positions = self.positions[:
                                                                         num_input_tokens]
                attn_metadata_i.prefill.max_seq_lens += 1
                attn_metadata_i.prefill.max_seq_lens = min(
                    attn_metadata_i.prefill.max_seq_lens,
                    self.runner.model_config.max_model_len)
            if attn_metadata_i.decode is not None:
                attn_metadata_i.decode.seq_lens = attn_metadata_i.seq_lens
                attn_metadata_i.decode.seq_lens_list = attn_metadata_i.decode.seq_lens.tolist(
                )
                attn_metadata_i.decode.input_positions = self.positions[:
                                                                        num_input_tokens]
                attn_metadata_i.decode.max_seq_lens += 1
                attn_metadata_i.decode.max_seq_lens = min(
                    attn_metadata_i.decode.max_seq_lens,
                    self.runner.model_config.max_model_len)

        # mtp>1: [batch_size, k]
        draft_token_ids = torch.stack(draft_token_ids_list, dim=1)
        return draft_token_ids

    def _get_torchair_lazy_compiled_model(self, batch_size: int):
        if batch_size < 0 or batch_size > self.runner.torchair_graph_batch_sizes[
                -1]:
            raise ValueError(
                f"Bad graph batch size:{batch_size}! max_graph_batch_sizes:{self.runner.torchair_graph_batch_sizes[-1]}"
            )

        compiled_model = self.torchair_compiled_models.get(
            batch_size
        ) if self.runner.use_cached_npu_graph else self.torchair_compiled_model

        if compiled_model:
            return compiled_model

        patch_for_hcom()
        config = torchair.CompilerConfig()
        config.experimental_config.frozen_parameter = True
        config.experimental_config.tiling_schedule_optimize = True
        config.experimental_config.enable_view_optimize = \
        get_ascend_config().torchair_graph_config.enable_view_optimize
        torch.npu.set_compile_mode(jit_compile=False)
        if not self.runner.use_cached_npu_graph:
            npu_backend = torchair.get_npu_backend(compiler_config=config)
            self.torchair_compiled_model = torch.compile(
                self.model,
                dynamic=not self.use_sparse,
                fullgraph=True,
                backend=npu_backend)
            return self.torchair_compiled_model
        else:
            # Generate a new forward proxy code object to prevent the invalidation of
            # compilation cache caused by dynamo retracing
            forward_proxy_name = f"{self.model.__class__.__name__}_forward_with_batch_size_{batch_size}"
            forward_fn = self.model.forward
            code = forward_fn.__code__
            # Mark code object with a new proxy name
            modified_code = code.replace(co_name=forward_proxy_name, )

            modified_func = types.FunctionType(modified_code,
                                               forward_fn.__globals__,
                                               name=forward_proxy_name,
                                               argdefs=forward_fn.__defaults__)

            self.model.__dict__[forward_proxy_name] = modified_func.__get__(
                self.model, nn.Module)
            self.torchair_compiled_models[
                batch_size] = torchair.inference.cache_compile(
                    self.model.__dict__[forward_proxy_name],
                    dynamic=not self.use_sparse,
                    fullgraph=True,
                    cache_dir=TORCHAIR_CACHE_DIR,
                    config=config,
                    ge_cache=False)
            return self.torchair_compiled_models[batch_size]

    # TODO Using torch instead of triton may result in poor performance
    def _prepare_input_kernel(self, out_ptr: torch.Tensor,
                              cu_query_lens: torch.Tensor,
                              cu_num_tokens: torch.Tensor, block_size: int):
        device = cu_query_lens.device
        dtype = out_ptr.dtype

        offsets = torch.arange(block_size, device=device, dtype=dtype)
        start_pos = cu_num_tokens[:-1]
        end_pos = cu_num_tokens[1:]
        num_tokens = end_pos - start_pos

        global_indices = (start_pos.view(-1, 1) + offsets.view(1, -1))
        values = (cu_query_lens[:-1].view(-1, 1) + offsets.view(1, -1))

        mask = (offsets.view(1, -1) < num_tokens.view(-1, 1))

        global_indices_flat = global_indices[mask]
        values_flat = values[mask]
        out_ptr[global_indices_flat] = values_flat



# Copyright Huawei Technologies Co., Ltd. 2024-2025. All rights reserved.
# Todo: Once https://github.com/vllm-project/vllm/pull/24069 is merged in vllm. Remove this policy.

import logging
from collections import deque
from typing import Dict

import numpy as np
import torch
from numba import njit  # type: ignore

from .policy_abstract import DynamicConfig, EplbPolicy

numba_logger = logging.getLogger("numba")
numba_logger.setLevel(logging.WARNING)


@njit
def compute_piece_counts(X, P, stage_weights):
    n_stage, N = X.shape
    S = P - N
    pieces = np.ones(N, dtype=np.int32)
    unit = X / pieces  # unit[i, j] = X[i, j] / pieces[j]

    for _ in range(S):
        deltas = np.zeros(N, dtype=np.float32)
        for i in range(n_stage):
            # Find top1 and top2
            idx1 = -1
            idx2 = -1
            val1 = -1.0
            val2 = -1.0
            for j in range(N):
                v = unit[i, j]
                if v > val1:
                    val2 = val1
                    idx2 = idx1
                    val1 = v
                    idx1 = j
                elif v > val2:
                    val2 = v
                    idx2 = j

            origin = unit[i, idx1]
            secv = unit[i, idx2]
            alt = X[i, idx1] / (pieces[idx1] + 1)
            delta = origin - (alt if alt > secv else secv)
            deltas[idx1] += delta * stage_weights[i] if np.any(
                delta) != 0 else stage_weights[i]

        max_idx = np.argmax(deltas)
        pieces[max_idx] += 1
        for i in range(n_stage):
            unit[i, max_idx] = X[i, max_idx] / pieces[max_idx]

    # Compute max load
    max_load = 0.0
    for j in range(N):
        total = 0.0
        for i in range(n_stage):
            total += unit[i, j]
        if total > max_load:
            max_load = total

    return pieces


@njit
def jsq_placement(X, pieces, M, stage_weights):
    n_stage, N = X.shape
    total_piece = pieces.sum()
    num_per_group = total_piece // M

    # 1. Compute unit_hotness
    unit_hotness = np.empty((n_stage, N), dtype=np.float32)
    for i in range(N):
        if pieces[i] > 0:
            for s in range(n_stage):
                unit_hotness[s, i] = X[s, i] / pieces[i]
        else:
            for s in range(n_stage):
                unit_hotness[s, i] = 0.0

    # 2. Sort by total hotness
    scores = np.zeros(N, dtype=np.float32)
    for i in range(N):
        for s in range(n_stage):
            scores[i] += unit_hotness[s, i]
    idx = np.argsort(-scores)

    # 3. Initialization
    loads = np.zeros((n_stage, M), dtype=np.float32)
    dev_phy_exp_n = np.zeros(M, dtype=np.int32)
    deployment = -np.ones((M, num_per_group), dtype=np.int32)
    dep_ptr = np.zeros(M, dtype=np.int32)

    # 4. Main loop
    for t in range(N):
        i = idx[t]
        used_device = list()
        for _ in range(pieces[i]):
            # 4.1 Construct w vector
            w = np.empty(n_stage, dtype=np.float32)
            for s in range(n_stage):
                w[s] = unit_hotness[s, i]

            # 4.2 Compute stage-level maximum load
            stage_max = np.empty(n_stage, dtype=np.float32)
            for s in range(n_stage):
                max_val = loads[s, 0]
                for k in range(1, M):
                    if loads[s, k] > max_val:
                        max_val = loads[s, k]
                stage_max[s] = max_val

            # 4.3 Compute denominator
            denom = np.empty(n_stage, dtype=np.float32)
            for s in range(n_stage):
                sum_tmp = 0.0
                for j in range(M):
                    sum_tmp += loads[s, j] + w[s]
                denom[s] = sum_tmp / M + 1e-2

            # 4.4 Find best device j
            best_j = -1
            best_val = 1e30
            for j in range(M):
                if dev_phy_exp_n[j] >= num_per_group:
                    continue
                if j in used_device:
                    continue
                score = 0.0
                for s in range(n_stage):
                    tmp_sj = loads[s, j] + w[s]
                    numer_sj = tmp_sj if tmp_sj > stage_max[s] else stage_max[s]
                    score += stage_weights[s] * (numer_sj / denom[s])
                if score < best_val:
                    best_val = score
                    best_j = j
            if best_j == -1:
                continue

            used_device.append(best_j)

            # 4.5 Update status
            for s in range(n_stage):
                loads[s, best_j] += w[s]
            ptr = dep_ptr[best_j]
            deployment[best_j, ptr] = i
            dep_ptr[best_j] += 1
            dev_phy_exp_n[best_j] += 1

    # Handle remaining -1 values: fill with random elements from range(N) not in current column
    for rank in range(M):
        for col in range(num_per_group):
            if deployment[rank, col] == -1:
                # Get elements already in current column
                current_rank_elements = set(deployment[rank, :])
                # Filter elements from range(N) not in current column
                available = [
                    x for x in range(N) if x not in current_rank_elements
                ]
                # Randomly select an available element to fill
                if len(available) > 0:
                    rand_idx = np.random.randint(0, len(available))
                    deployment[rank, col] = available[rand_idx]
                elif N > 0:
                    # All unique experts are already in this rank's column, so we can pick any expert randomly.
                    deployment[rank, col] = np.random.randint(0, N)

    return deployment


@njit
def slice_values(X, pieces):
    total_len = 0
    for i in range(X.shape[0]):
        total_len += pieces[i]
    result = np.empty(total_len, dtype=np.float32)
    idx = 0
    for i in range(X.shape[0]):
        val = X[i] / pieces[i]
        for _ in range(pieces[i]):
            result[idx] = val
            idx += 1
    return result


@njit
def group_based_adaptive_bloating_kernel(X, P, M, simulated_pieces,
                                         simulated_deployment, stage_weights):
    n_stage, N = X.shape
    num_group = P // M

    X_all = np.zeros(N, dtype=np.float32)
    for i in range(n_stage):
        for j in range(N):
            X_all[j] += X[i, j]

    sort_idx = np.argsort(np.negative(X_all))
    X_sorted = X[:, sort_idx]

    unit_load = np.empty(N, dtype=np.float32)
    for j in range(N):
        unit_load[j] = X_all[j] / simulated_pieces[j]

    flat_deployment = simulated_deployment.reshape(-1)
    simulated_load = np.zeros(M, dtype=np.float32)
    for i in range(flat_deployment.shape[0]):
        simulated_load[i // (flat_deployment.shape[0] //
                             M)] += unit_load[flat_deployment[i]]

    slice_vals = slice_values(X_all, simulated_pieces)
    sorted_slices = np.sort(slice_vals)[::-1]
    simulated_slopes = (sorted_slices[:-M + 1] - sorted_slices[M - 1:]) / M

    cumulative_slices_used = np.zeros(N, dtype=np.int32)
    acc = 0
    for i in range(N):
        acc += simulated_pieces[sort_idx[i]]
        cumulative_slices_used[i] = acc

    group_boundary_indices = np.zeros(num_group, dtype=np.int32)
    for i in range(1, num_group + 1):
        for j in range(N):
            if cumulative_slices_used[j] >= i * M:
                group_boundary_indices[i - 1] = j
                break

    slices_used_per_group = np.zeros(num_group, dtype=np.int32)
    slices_used_per_group[0] = group_boundary_indices[0]
    for i in range(1, num_group):
        slices_used_per_group[
            i] = group_boundary_indices[i] - group_boundary_indices[i - 1]
    slices_used_per_group = M - slices_used_per_group

    loads = np.zeros(M, dtype=np.float32)
    pieces = np.zeros(N, dtype=np.int32)
    num_remain_slice = P - N
    current_idx = 0

    for g in range(num_group):
        window = X_sorted[:, current_idx:current_idx + 2 * M]
        low = max(0, current_idx + M - N)
        high = min(num_remain_slice, M - 1)

        while (high - low) > 1:
            mid = int((high + low) // 2)
            keep = M - mid
            current_group = window[:, :keep]
            current_pieces = compute_piece_counts(current_group, M,
                                                  stage_weights)
            current_pieces = np.maximum(current_pieces, 1)
            current_slice = slice_values(current_group.sum(0), current_pieces)
            current_slice_sorted = np.sort(current_slice)
            current_loads = loads + current_slice_sorted
            current_max: np.float32 = np.max(current_loads)
            current_min: np.float32 = np.min(current_loads)
            current_slope = (current_max - current_min) / M
            next_slope: np.float32 = np.max(simulated_slopes[current_idx +
                                                             keep:])

            if abs(current_slope) > abs(next_slope):
                low = mid
            else:
                high = mid

        S = high
        keep = M - S
        current_group = window[:, :keep]
        current_pieces = compute_piece_counts(current_group, M, stage_weights)

        for i in range(keep):
            pieces[sort_idx[current_idx + i]] = current_pieces[i]

        current_slice = slice_values(current_group.sum(0), current_pieces)
        current_slice_sorted = np.sort(current_slice)
        loads += current_slice_sorted
        loads = np.sort(loads)[::-1]

        current_idx += keep
        num_remain_slice -= S

    return pieces


@njit
def compute_objective(deployment, X, pieces):
    M, P = deployment.shape
    loads = np.zeros(M)

    for i in range(M):
        for j in range(P):
            expert = deployment[i, j]
            if pieces[expert] == 0:
                continue
            loads[i] += X[expert] / pieces[expert]

    mean_load = np.mean(loads)
    max_load: np.float32 = np.max(loads)
    obj = max_load / mean_load
    return obj, loads


@njit
def auto_fix_new_placement(old_placement, new_placement):
    """
    Adjust the new_placement matrix to ensure elements (including duplicates) that exist in both
    old_placement and new_placement remain in their original positions from old_placement.
    New elements (unique to new_placement) will fill the remaining empty positions.

    Args:
        old_placement: Old deployment matrix with shape (num_ranks, num_experts)
        new_placement: New deployment matrix to be fixed, must have the same shape as old_placement

    Returns:
        fixed_new: adjusted version of the new_placement matrix
    """
    num_ranks, num_experts = old_placement.shape
    fixed_new = np.empty_like(new_placement)

    max_expert_old = old_placement.max() if num_experts > 0 else 0
    max_expert_new = new_placement.max() if num_experts > 0 else 0
    max_expert = max(max_expert_old, max_expert_new)

    for rank_id in range(num_ranks):
        old_row = old_placement[rank_id]
        new_row = new_placement[rank_id]

        index_array = np.full((max_expert + 1, num_experts),
                              -1,
                              dtype=np.int32)
        count_array = np.zeros(max_expert + 1, dtype=np.int32)

        for idx in range(num_experts):
            val = old_row[idx]
            if val >= 0 and val <= max_expert:
                pos = count_array[val]
                index_array[val, pos] = idx
                count_array[val] += 1

        old_counter = np.zeros(max_expert + 1, dtype=np.int32)
        for idx in range(num_experts):
            val = old_row[idx]
            if val >= 0 and val <= max_expert:
                old_counter[val] += 1

        retain_elements = np.empty(num_experts, dtype=new_placement.dtype)
        new_elements = np.empty(num_experts, dtype=new_placement.dtype)
        retain_ptr = 0
        new_ptr = 0

        for val in new_row:
            if val >= 0 and val <= max_expert and old_counter[val] > 0:
                retain_elements[retain_ptr] = val
                retain_ptr += 1
                old_counter[val] -= 1
            else:
                new_elements[new_ptr] = val
                new_ptr += 1

        current_fixed = np.full(num_experts, -1, dtype=new_placement.dtype)

        for i in range(retain_ptr):
            val = retain_elements[i]
            if val >= 0 and val <= max_expert:
                pos = count_array[val] - 1
                if pos >= 0:
                    idx = index_array[val, pos]
                    current_fixed[idx] = val
                    count_array[val] -= 1

        empty_indices = np.empty(num_experts, dtype=np.int32)
        empty_ptr = 0
        for idx in range(num_experts):
            if current_fixed[idx] == -1:
                empty_indices[empty_ptr] = idx
                empty_ptr += 1

        for i in range(new_ptr):
            if i < empty_ptr:
                current_fixed[empty_indices[i]] = new_elements[i]

        fixed_new[rank_id] = current_fixed

    return fixed_new


class FlashLB(EplbPolicy):

    def __init__(self, config: DynamicConfig):
        super().__init__(config)
        self.par_history: Dict[int, float] = {}
        self.hotness_window: Dict[int, deque[float]] = {}
        self.max_stage_window = (config.max_stage_window if hasattr(
            config, "max_stage_window") else 1)
        self.buffer_expert_layer_num = (
            config.buffer_expert_layer_num if hasattr(
                config, "buffer_expert_layer_num") else 58)
        self.threshold_ratio = (config.threshold_ratio if hasattr(
            config, "threshold_ratio") else 0)

    def compute_expert_hotness(self, num_of_expert: int,
                               deployment: np.ndarray, rank_load: np.ndarray):
        hotness = np.zeros(num_of_expert, dtype=rank_load.dtype)
        deployment_flat = deployment.ravel()
        rank_load_flat = rank_load.ravel()
        np.add.at(hotness, deployment_flat, rank_load_flat)
        return hotness

    def compute_rank_load(self, deployment: np.ndarray, hotness: np.ndarray):
        n_stage, N = hotness.shape
        if np.any(deployment < 0):
            print(f"Invalid deployment with negative values: {deployment}")
            raise ValueError("Deployment table contains negative values.")
        counts = np.bincount(deployment.reshape(-1), minlength=N)
        unit_hotness = np.divide(hotness,
                                 counts,
                                 out=np.zeros_like(hotness, dtype=float),
                                 where=counts != 0)
        stage_par = np.zeros(n_stage)
        for i in range(n_stage):
            stage_load = unit_hotness[i][deployment].sum(-1)
            stage_par[i] = stage_load.max() / stage_load.mean()
        return stage_par.mean()

    def group_based_adaptive_bloating(self,
                                      X,
                                      P,
                                      M,
                                      stage_weights=None,
                                      recorsive=False):
        n_stage, N = X.shape
        if stage_weights is None:
            stage_weights = np.ones(n_stage, dtype=np.float32)

        if recorsive:
            (
                simulated_deployment,
                simulated_pieces,
            ) = self.group_based_adaptive_bloating(X,
                                                   P,
                                                   M,
                                                   stage_weights,
                                                   recorsive=False)
        else:
            simulated_pieces = compute_piece_counts(X, P, stage_weights)
            simulated_deployment = jsq_placement(X, simulated_pieces, M,
                                                 stage_weights)

        pieces = group_based_adaptive_bloating_kernel(
            X.astype(np.float32),
            P,
            M,
            simulated_pieces.astype(np.int32),
            simulated_deployment.astype(np.int32),
            stage_weights.astype(np.float32),
        )

        deployment = jsq_placement(X, pieces, M, stage_weights)

        X_all = X.sum(0)
        unit_load = np.divide(X_all,
                              pieces,
                              out=np.zeros_like(X_all, dtype=float),
                              where=pieces != 0)
        load = unit_load[deployment].sum(-1)

        sim_unit_load = X_all / simulated_pieces
        sim_load = sim_unit_load[simulated_deployment].sum(-1)

        if load.max() > sim_load.max():
            return simulated_deployment, simulated_pieces
        return deployment, pieces

    def need_update(self, current_par, layer_id=0):
        threshold = self.par_history.get(layer_id, 0.0)
        return current_par >= self.threshold_ratio * threshold

    def compute_stage_weight(self, hotness):
        n_stage = hotness.shape[0]
        stage_weights = np.zeros(n_stage)
        for i in range(n_stage):
            stage_weights[i] = hotness[i].sum()

        stage_weights = stage_weights / stage_weights.max()
        return stage_weights

    def rebalance_layer(self, deployment, hotness, layer_id=0):
        num_rank, expert_per_rank = deployment.shape
        num_expert = np.unique(deployment.reshape(-1)).shape[0]
        num_of_redundant_expert = num_rank * expert_per_rank - num_expert

        current_par = self.compute_rank_load(deployment, hotness)

        if not self.need_update(current_par, layer_id):
            return deployment, current_par, current_par

        stage_weights = self.compute_stage_weight(hotness)
        new_deployment, _ = self.group_based_adaptive_bloating(
            hotness,
            num_expert + num_of_redundant_expert,
            num_rank,
            stage_weights,
            recorsive=False,
        )
        if np.any(new_deployment < 0):
            print(f"{new_deployment=}")
        new_par = self.compute_rank_load(new_deployment, hotness)

        return new_deployment, new_par, current_par

    def register_hotness(self, deployment, rank_load, num_layer, num_expert):
        for layer in range(num_layer):
            if layer not in self.hotness_window:
                self.hotness_window[layer] = deque(
                    maxlen=self.max_stage_window)
            hotness = self.compute_expert_hotness(num_expert,
                                                  deployment[layer],
                                                  rank_load[layer])
            self.hotness_window[layer].append(hotness)

    def compress_by_avg_pooling_fast_nd(self, arr, m):
        n, d = arr.shape
        idx = (np.arange(n) * m // n)
        result = np.zeros((m, d))
        counts = np.zeros((m, 1))
        np.add.at(result, idx, arr)
        np.add.at(counts, idx, 1)
        return result / counts

    def rebalance_experts(self, current_expert_table, expert_workload):
        current_deployment = np.array(current_expert_table)
        expert_workload = np.array(expert_workload)
        expert_workload += 1
        num_layer = expert_workload.shape[0]
        num_expert = np.unique(current_expert_table[0].reshape(-1)).shape[0]
        self.register_hotness(current_deployment, expert_workload, num_layer,
                              num_expert)

        new_deployment = current_deployment.copy()

        layers_need_update = np.arange(num_layer)

        new_par = np.zeros(layers_need_update.shape[0])
        current_par = np.zeros(layers_need_update.shape[0])
        for i, layer in enumerate(layers_need_update):
            hotness = np.array(self.hotness_window[layer])
            if hotness.shape[0] > self.max_stage_window:
                hotness = self.compress_by_avg_pooling_fast_nd(
                    hotness, self.max_stage_window)

            (
                new_deployment[layer],
                new_par[i],
                current_par[i],
            ) = self.rebalance_layer(current_deployment[layer],
                                     hotness,
                                     layer_id=layer)

        priority = new_par / current_par
        priority_idx = np.argsort(priority)
        priority_idx = priority_idx[priority[priority_idx] <
                                    1][:self.buffer_expert_layer_num]

        if np.all(expert_workload == 1):
            for _, layer in enumerate(layers_need_update):
                self.hotness_window[layer].pop()
            return False, np.array([], dtype=int), current_deployment
        change = len(priority_idx) > 0
        if change:
            for idx in priority_idx:
                self.par_history[layers_need_update[idx]] = new_par[idx]

        layers_need_update = priority_idx
        deployment = current_deployment
        for layer in layers_need_update:
            deployment[layer] = auto_fix_new_placement(
                current_deployment[layer], new_deployment[layer])

        return change, layers_need_update, deployment


def generate_layered_experts(num_layers=58,
                             layer_shape=(32, 9),
                             expert_min=0,
                             expert_max=255):
    """
    Generate expert deployment matrix meeting the following conditions:
    - Total of num_layers layers
    - Each layer has shape layer_shape (32,9)
    - Each expert from expert_min to expert_max (0 to 255) appears at least once in each layer

    Args:
        num_layers: Number of layers, default 58
        layer_shape: Shape of a single layer, default (32,9)
        expert_min: Minimum expert ID, default 0
        expert_max: Maximum expert ID, default 255
    Returns:
        torch.Tensor: Tensor with shape (num_layers, layer_shape[0], layer_shape[1])
    """
    # 1. Basic parameter calculation
    expert_num = expert_max - expert_min + 1  # Total number of experts: 256 (0~255)
    layer_total = layer_shape[0] * layer_shape[
        1]  # Total elements in a single layer: 32*9=288
    extra_slots = layer_total - expert_num  # Number of random positions to fill per layer: 288-256=32

    # 2. Verify feasibility (total elements must be ≥ number of experts to cover all experts)
    assert layer_total >= expert_num, (
        f"Number of elements in a single layer {layer_total} < number of experts {expert_num}, "
        "cannot cover all experts")

    # 3. Generate layers one by one
    layers = []
    for _ in range(num_layers):
        # 3.1 Generate "complete expert sequence" (ensure each expert from 0 to 255 is included)
        full_experts = torch.arange(expert_min,
                                    expert_max + 1,
                                    dtype=torch.int64)  # shape (256,)

        # 3.2 Generate "supplementary random experts" (fill remaining 32 positions, randomly selected from 0~255)
        extra_experts = torch.randint(expert_min,
                                      expert_max + 1,
                                      size=(extra_slots, ),
                                      dtype=torch.int64)  # shape (32,)

        # 3.3 Concatenate and shuffle (ensure random distribution of experts in each layer)
        layer_flat = torch.cat([full_experts, extra_experts],
                               dim=0)  # shape (288,)
        # Shuffle order (use randperm to generate random indices to avoid repeated shuffling issues)
        shuffle_idx = torch.randperm(layer_flat.shape[0])
        layer_shuffled = layer_flat[shuffle_idx]

        # 3.4 Reshape to layer_shape (32,9)
        layer = layer_shuffled.reshape(layer_shape)
        layers.append(layer)

    # 4. Stack all layers to get the final tensor
    return torch.stack(layers, dim=0)  # shape (58,32,9)


def warm_up():
    exam_config = DynamicConfig()
    exam_config.ep_worldsize = 32
    exam_config.num_die_per_host = 16
    algo = FlashLB(exam_config)
    # Generate target tensor
    expert_tensor = generate_layered_experts(num_layers=58,
                                             layer_shape=(32, 9))

    algo.rebalance_experts(expert_tensor, torch.randint(1, 1000, (58, 32, 9)))



#
# Copyright (c) 2025 Huawei Technologies Co., Ltd. All Rights Reserved.
# Copyright 2023 The vLLM team.
#
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# This file is a part of the vllm-ascend project.

import torch
import vllm
from vllm.model_executor.models.utils import (_embedding_count_expression,
                                              _flatten_embeddings)
from vllm.multimodal import NestedTensors


def _merge_multimodal_embeddings(
    inputs_embeds: torch.Tensor,
    is_multimodal: torch.Tensor,
    multimodal_embeddings: NestedTensors,
) -> torch.Tensor:
    """
    Merge ``multimodal_embeddings`` into ``inputs_embeds`` by overwriting the
    positions in ``inputs_embeds`` corresponding to placeholder tokens in
    ``input_ids``.

    Note:
        This updates ``inputs_embeds`` in place.
    """
    flattened = _flatten_embeddings(multimodal_embeddings)
    try:
        inputs_embeds[is_multimodal] = flattened
    except RuntimeError as e:
        num_expected_tokens = is_multimodal.sum().item()
        assert isinstance(num_expected_tokens, int)

        if flattened.shape[0] != num_expected_tokens:
            expr = _embedding_count_expression(multimodal_embeddings)
            raise ValueError(
                f"Attempted to assign {expr} = {flattened.shape[0]} "
                f"multimodal tokens to {num_expected_tokens} placeholders"
            ) from e
        else:
            raise ValueError("Error during masked scatter operation") from e

    return inputs_embeds


vllm.model_executor.models.utils._merge_multimodal_embeddings = _merge_multimodal_embeddings



#
# Copyright (c) 2025 Huawei Technologies Co., Ltd. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch


def bgmv_shrink(inputs: torch.Tensor,
                lora_a_weights: torch.Tensor,
                output_tensor: torch.Tensor,
                lora_indices_tensor: torch.Tensor,
                scaling: float = 1.0):
    return torch.ops._C_ascend.bgmv_shrink(
        inputs,
        lora_a_weights,
        lora_indices_tensor,
        output_tensor,
        scaling,
    )


def bgmv_expand(inputs: torch.Tensor,
                lora_b_weights: torch.Tensor,
                output_tensor: torch.Tensor,
                lora_indices_tensor: torch.Tensor,
                add_inputs: bool = True):
    return torch.ops._C_ascend.bgmv_expand(
        inputs,
        lora_b_weights,
        lora_indices_tensor,
        output_tensor,
        0,
        output_tensor.size(1),
    )


def bgmv_expand_slice(inputs: torch.Tensor,
                      lora_b_weights: torch.Tensor,
                      output_tensor: torch.Tensor,
                      lora_indices_tensor: torch.Tensor,
                      slice_offset: int,
                      slice_size: int,
                      add_inputs: bool = True):
    return torch.ops._C_ascend.bgmv_expand(inputs, lora_b_weights,
                                           lora_indices_tensor, output_tensor,
                                           slice_offset, slice_size)


def sgmv_shrink(
    inputs: torch.Tensor,
    lora_a_weights: torch.Tensor,
    output_tensor: torch.Tensor,
    b_seq_start_loc: torch.Tensor,
    seq_len_tensor: torch.Tensor,
    lora_indices_tensor: torch.Tensor,
    batches: int,
    max_seq_length: int,
    token_nums: int,
    scaling: float,
):
    return torch.ops._C_ascend.sgmv_shrink(inputs, lora_a_weights,
                                           lora_indices_tensor, seq_len_tensor,
                                           output_tensor, scaling)


def sgmv_expand(inputs: torch.Tensor,
                lora_b_weights: torch.Tensor,
                output_tensor: torch.Tensor,
                b_seq_start_loc: torch.Tensor,
                seq_len_tensor: torch.Tensor,
                lora_indices_tensor: torch.Tensor,
                batches: int,
                max_seq_length: int,
                token_nums: int,
                add_inputs: bool = False):
    return torch.ops._C_ascend.sgmv_expand(
        inputs,
        lora_b_weights,
        lora_indices_tensor,
        seq_len_tensor,
        output_tensor,
        0,
        output_tensor.size(1),
    )


def sgmv_expand_slice(inputs: torch.Tensor,
                      lora_b_weights: torch.Tensor,
                      output_tensor: torch.Tensor,
                      b_seq_start_loc: torch.Tensor,
                      seq_len_tensor: torch.Tensor,
                      lora_indices_tensor: torch.Tensor,
                      batches: int,
                      max_seq_length: int,
                      token_nums: int,
                      slice_offset: int,
                      slice_size: int,
                      add_inputs: bool = False):
    return torch.ops._C_ascend.sgmv_expand(inputs, lora_b_weights,
                                           lora_indices_tensor, seq_len_tensor,
                                           output_tensor, slice_offset,
                                           slice_size)



#
# Copyright (c) 2025 Huawei Technologies Co., Ltd. All Rights Reserved.
# This file is a part of the vllm-ascend project.
# Adapted from vllm-project/vllm/tests/utils.py
# Copyright 2023 The vLLM team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import functools
import os
import signal
from collections.abc import Sequence
from typing import Callable

import torch
import torch.nn.functional as F
from typing_extensions import ParamSpec

_P = ParamSpec("_P")


def fork_new_process_for_each_test(
        f: Callable[_P, None]) -> Callable[_P, None]:
    """Decorator to fork a new process for each test function.
    See https://github.com/vllm-project/vllm/issues/7053 for more details.
    """

    @functools.wraps(f)
    def wrapper(*args: _P.args, **kwargs: _P.kwargs) -> None:
        # Make the process the leader of its own process group
        # to avoid sending SIGTERM to the parent process
        os.setpgrp()
        from _pytest.outcomes import Skipped
        pid = os.fork()
        print(f"Fork a new process to run a test {pid}")
        if pid == 0:
            try:
                f(*args, **kwargs)
            except Skipped as e:
                # convert Skipped to exit code 0
                print(str(e))
                os._exit(0)
            except Exception:
                import traceback
                traceback.print_exc()
                os._exit(1)
            else:
                os._exit(0)
        else:
            pgid = os.getpgid(pid)
            _pid, _exitcode = os.waitpid(pid, 0)
            # ignore SIGTERM signal itself
            old_signal_handler = signal.signal(signal.SIGTERM, signal.SIG_IGN)
            # kill all child processes
            os.killpg(pgid, signal.SIGTERM)
            # restore the signal handler
            signal.signal(signal.SIGTERM, old_signal_handler)
            assert _exitcode == 0, (f"function {f} failed when called with"
                                    f" args {args} and kwargs {kwargs}")

    return wrapper


def matryoshka_fy(tensor: torch.Tensor, dimensions: int):
    tensor = torch.tensor(tensor)
    tensor = tensor[..., :dimensions]
    tensor = F.normalize(tensor, p=2, dim=1)
    return tensor


def check_embeddings_close(
    *,
    embeddings_0_lst: Sequence[list[float]],
    embeddings_1_lst: Sequence[list[float]],
    name_0: str,
    name_1: str,
    tol: float = 1e-3,
) -> None:
    assert len(embeddings_0_lst) == len(embeddings_1_lst)

    for prompt_idx, (embeddings_0, embeddings_1) in enumerate(
            zip(embeddings_0_lst, embeddings_1_lst)):
        assert len(embeddings_0) == len(embeddings_1), (
            f"Length mismatch: {len(embeddings_0)} vs. {len(embeddings_1)}")

        sim = F.cosine_similarity(torch.tensor(embeddings_0),
                                  torch.tensor(embeddings_1),
                                  dim=0)

        fail_msg = (f"Test{prompt_idx}:"
                    f"\nCosine similarity: \t{sim:.4f}"
                    f"\n{name_0}:\t{embeddings_0[:16]!r}"
                    f"\n{name_1}:\t{embeddings_1[:16]!r}")

        assert sim >= 1 - tol, fail_msg



#
# Copyright (c) 2025 Huawei Technologies Co., Ltd. All Rights Reserved.
# This file is a part of the vllm-ascend project.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# ----------------------------------------------------------------------------------
# This module manage the patch for vllm. There are two folders in this module:
# - platform: contains the patches applied before worker starts. It's called by
#             `vllm_ascend.utils.adapt_patch(is_global_patch=True)` in
#             `vllm_ascend.platform.NPUPlatform.pre_register_and_update()` function.
# - worker: contains the patches applied when worker starts. It's called by
#           `vllm_ascend.utils.adapt_patch(is_global_patch=False)` in
#           each worker's `__init__` function.
#
# Once a new patch is added in vllm-ascend, please add the patch description into this file as well.
# ----------------------------------------------------------------------------------

# What's Patched and how it works:
# --------------------------------
# * Platform Patch:
# =================
# ** File: platform/patch_distributed.py**
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#   1. `vllm.config.ParallelConfig.get_next_dp_init_port`
#    Why:
#       vllm doesn't support get port from environment.
#    How：
#       Add the logic to get port from environment.
#    Related PR (if no, explain why):
#       Need a PR to vllm to support get port from environment.
#    Future Plan:
#       Remove those patch when vllm merged them
#   2. `torch.distributed.all_reduce`, `torch.distributed.broadcast`
#    Why:
#       tensor alignment for 310p
#    How：
#       rewrite all_reduce and broadcast in torch.distributed
#    Related PR (if no, explain why):
#       No, not ready yet.
#    Future Plan:
#       Find a better way to support tensor alignment for 310p without this patch.
#
# ** File: worker/patch_multimodal_merge.py**
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#   1. `vllm.model_executor.models.utils._merge_multimodal_embeddings`
#    Why:
#       '_merge_multimodal_embeddings' func of vllm is incompatible with Ascend.
#    How：
#       Replace with CPU operation that can be executed asynchronously.
#    Related PR (if no, explain why):
#       This is a bug by Ascend only. It can' be fixed in vLLM.
#    Future Plan:
#       Identify this pattern in torch-npu and remove this patch.
#
# * Worker Patch:
# ===============
# ** File: worker/patch_minicpm.py **
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#   1. `vllm.model_executor.models.minicpm.MiniCPMAttention.forward`
#    Why:
#       The forward func of MiniCPMAttention in vllm do a datatype convert
#       (original datatype --> float32) to ensure the precision on cuda.
#       However float32 is not supported in cann rope op, thus we keep this patch
#    How：
#       Removed the dtype convert operations in forward
#    Related PR (if no, explain why):
#       NO, only for npu due to rope op.
#    Future Plan:
#       Keep this patch in vllm-ascend.
#
# ** File: worker/patch_distributed.py **
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#   1. `vllm.distributed.parallel_state.GroupCoordinator`
#   (1) __init__()
#    Why:
#       The original GroupCoordinator initialization lacks pg_options to generate new
#       process group with customized options.
#    How:
#       Inject HCCL options during process group initialization.
#    Related PR (if no, explain why):
#       Need a PR to vllm to support a dictionary as input while initializing distributed
#       environment (e.g., Dict[str, torch.distributed.ProcessGroupHCCL.Options])
#       https://github.com/vllm-project/vllm/pull/25417
#    Future Plan:
#       Remove this patch when vllm merges this PR.
#   (2) all_to_all()
#    Why:
#       vllm doesn't support all_to_all for GroupCoordinator.
#    How：
#       Add all_to_all implementation for GroupCoordinator.
#    Related PR (if no, explain why):
#       Need a PR to vllm to support all_to_all for GroupCoordinator.
#    Future Plan:
#       Remove this patch when vllm merged them.
#
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#   1. `vllm.v1.sample.sampler.Sampler.gather_logprobs`
#    Why:
#       We need to patch gather_logprobs to make sure call batched_count_greater_than
#       with backend=current_platform.simple_compile_backend
#    How：
#       Patch gather_logprobs call new batched_count_greater_than
#    Related PR (if no, explain why):
#       - https://github.com/vllm-project/vllm/pull/21591
#    Future Plan:
#       Revert it when vLLM merge #21591 and release new version
# ** File: worker/patch_logits.py **
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#   1. `vllm._custom_ops.apply_repetition_penalties`
#    Why:
#       apply_repetition_penalties in vLLM use tensor.is_cuda to check if tensor is on cuda. But the value is always True
#       on ascend, thus we need to patch apply_repetition_penalties.
#    How：
#       Remove the related cuda check in apply_repetition_penalties.
#    Related PR (if no, explain why):
#       - this is a bug by Ascend only. It can' be fixed in vLLM.
#    Future Plan:
#       Fix this bug in torch-npu, bump torch-npu version and remove this patch.
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#   1. `vllm.model_executor.models.roberta.RobertaEmbedding.forward`
#    Why:
#       shift operation in `_encode_token_type_ids` and `_decode_token_type_ids` cannot run in ascend aclgraph mode
#    How：
#       Replace shift operation with multiplication and division.
#    Related PR (if no, explain why):
#       No, this need CANN add an aclnn shift operation
#    Future Plan:
#       Revert this when CANN support shift aclnn operation
#   2. `vllm.model_executor.models.roberta.RobertaForSequenceClassification.forward `
#    Why:
#       shift operation in `_encode_token_type_ids` and `_decode_token_type_ids` cannot run in ascend aclgraph mode
#    How：
#       Replace shift operation with multiplication and division.
#    Related PR (if no, explain why):
#       No, this need CANN add an aclnn shift operation
#    Future Plan:
#       Revert this when CANN support shift aclnn operation
#
# ** File: worker/patch_deepseek_mtp.py**
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#   1. `vllm.model_executor.models.deepseek_mtp.DeepSeekMultiTokenPredictorLayer.__init__`
#    Why:
#       '__init__' func of DeepSeekMultiTokenPredictorLayer didn't pass prefix to SharedHead.
#    How：
#       Replace with a new __init__.
#       Use a new SharedHead which passes prefix to ParallelLMHead.
#    Related PR (if no, explain why):
#       https://github.com/vllm-project/vllm/pull/25805
#    Future Plan:
#       Remove this patch when adapted vllm version contains the above PR.
#
# ** File: worker/patch_attention_layer.py **
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#   1. `vllm.attention.layer.Attention.forward`
#    Why:
#       There is a zerolike operator before the attention operation in each decoding stage.
#    How
#       Replace this zerolike operator with torch.empty
#    Related PR (if no, explain why):
#       - https://github.com/vllm-project/vllm/pull/26680
#    Future Plan:
#       Remove this to match the optimization supported in the VLLM version.
#



import gc

import torch

from vllm_ascend.utils import enable_custom_op

enable_custom_op()

DEFAULT_ATOL = 1e-3
DEFAULT_RTOL = 1e-3


def bgmv_shrink_cpu_impl(x: torch.Tensor, w: torch.Tensor,
                         indices: torch.Tensor, y: torch.tensor,
                         scaling: float) -> torch.Tensor:
    W = w[indices, :, :].transpose(-1, -2).to(torch.float32)
    z = torch.bmm(x.unsqueeze(1).to(torch.float32), W).squeeze()
    y[:, :] += z * scaling
    return y


@torch.inference_mode()
def test_bgmv_shrink():
    B = 1
    x = torch.randn([B, 128], dtype=torch.float16)
    w = torch.randn([64, 16, 128], dtype=torch.float16)
    indices = torch.zeros([B], dtype=torch.int64)
    y = torch.zeros([B, 16])

    x_npu = x.npu()
    w_npu = w.npu()
    indices_npu = indices.npu()
    y_npu = y.npu()

    y = bgmv_shrink_cpu_impl(x, w, indices, y, 0.5)
    torch.ops._C_ascend.bgmv_shrink(x_npu, w_npu, indices_npu, y_npu, 0.5)

    # Compare the results.
    torch.testing.assert_close(y_npu.cpu(),
                               y,
                               atol=DEFAULT_ATOL,
                               rtol=DEFAULT_RTOL)
    gc.collect()
    torch.npu.empty_cache()
    torch.npu.reset_peak_memory_stats()



#
# Copyright (c) 2025 Huawei Technologies Co., Ltd. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# This file is a part of the vllm-ascend project.
#

from typing import Optional, Tuple, Union

import torch
from vllm.config import get_current_vllm_config
from vllm.model_executor.layers.layernorm import RMSNorm

_original_re_init = RMSNorm.__init__


def torchair_rmsnorm_init_(
    self,
    hidden_size: int,
    eps: float = 1e-6,
    var_hidden_size: Optional[int] = None,
    has_weight: bool = True,
    dtype: Optional[torch.dtype] = None,
) -> None:
    _original_re_init(self, hidden_size, eps, var_hidden_size, has_weight,
                      dtype)
    vllm_config = get_current_vllm_config()
    self.bias = None
    # quantization with anti_method m4 will generate none-zero norm bias
    if vllm_config.quant_config is not None and \
            any("norm.bias" in name for name in vllm_config.quant_config.quant_description.keys()):
        self.bias = torch.nn.Parameter(torch.zeros(hidden_size),
                                       requires_grad=False)


def torchair_rmsnorm_forward_oot(
    self,
    x: torch.Tensor,
    residual: Optional[torch.Tensor] = None,
) -> Union[torch.Tensor, Tuple[torch.Tensor, torch.Tensor]]:
    """AscendRMSNorm forward in torchair mode.

    The key difference from the original implementation is the removal of operators
    from the torch.ops.vllm class, as these operators only function in non-torchair
    modes. Adding them back would cause the graph compilation to fail.
    """

    import torch_npu

    from vllm_ascend.utils import is_310p
    if residual is not None:
        if is_310p():
            orig_dtype = residual.dtype
            x = x + residual.to(x.dtype)
            residual = x.to(orig_dtype)
            x, _ = torch_npu.npu_rms_norm(x, self.weight,
                                          self.variance_epsilon)
        else:
            x, _, residual = torch_npu.npu_add_rms_norm(
                x, residual, self.weight, self.variance_epsilon)
        if self.bias is not None:
            x.add_(self.bias)
        return x, residual

    x, residual = torch_npu.npu_rms_norm(x, self.weight, self.variance_epsilon)
    if self.bias is not None:
        x.add_(self.bias)
    return x



#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# This file is a part of the vllm-ascend project.
#

from unittest.mock import MagicMock

import torch

from tests.ut.base import TestBase
from vllm_ascend.patch.worker.patch_minicpm import forward


class TestPatchMiniCPM(TestBase):

    def setUp(self):
        self.mock_self = MagicMock()

        self.mock_self.q_size = 128
        self.mock_self.kv_size = 128

        self.mock_self.qkv_proj = MagicMock()
        self.mock_self.rotary_emb = MagicMock()
        self.mock_self.attn = MagicMock()
        self.mock_self.o_proj = MagicMock()

        self.positions = torch.tensor([1, 2, 3])
        self.hidden_states = torch.randn(3, 256)

        self.mock_qkv = torch.randn(3, 384)
        self.mock_q = self.mock_qkv[:, :128]
        self.mock_k = self.mock_qkv[:, 128:256]
        self.mock_v = self.mock_qkv[:, 256:]

        self.mock_self.qkv_proj.return_value = (self.mock_qkv, None)
        self.mock_self.rotary_emb.return_value = (self.mock_q, self.mock_k)
        self.mock_self.attn.return_value = torch.randn(3, 256)
        self.mock_self.o_proj.return_value = (torch.randn(3, 256), None)

    def test_forward_patched(self):
        from vllm.model_executor.models.minicpm import MiniCPMAttention

        self.assertIs(MiniCPMAttention.forward, forward)

    def test_forward_function(self):
        result = forward(self.mock_self, self.positions, self.hidden_states)

        self.mock_self.qkv_proj.assert_called_once_with(self.hidden_states)

        args, _ = self.mock_self.rotary_emb.call_args
        self.assertEqual(len(args), 3)
        self.assertTrue(torch.equal(args[0], self.positions))
        self.assertTrue(torch.equal(args[1], self.mock_q))
        self.assertTrue(torch.equal(args[2], self.mock_k))

        args, _ = self.mock_self.attn.call_args
        self.assertEqual(len(args), 3)
        self.assertTrue(torch.equal(args[0], self.mock_q))
        self.assertTrue(torch.equal(args[1], self.mock_k))
        self.assertTrue(torch.equal(args[2], self.mock_v))

        self.mock_self.o_proj.assert_called_once_with(
            self.mock_self.attn.return_value)

        self.assertEqual(result.shape, (3, 256))
        self.assertTrue(
            torch.equal(result, self.mock_self.o_proj.return_value[0]))
